import os
from pathlib import Path
from textwrap import dedent

import pytest

from osa.configs import options
from osa.configs.config import cfg, DEFAULT_CFG

extra_files = Path(os.getenv("OSA_TEST_DATA", "extra"))
datasequence_history_file = extra_files / "history_files/sequence_LST1_04185.0010.history"
calibration_history_file = extra_files / "history_files/sequence_LST1_04183.history"
options.date = "2020-01-17"
options.tel_id = "LST1"
options.prod_id = "v0.1.0"


def test_historylevel():
    from osa.job import historylevel

    options.dl1_prod_id = "tailcut84"
    options.dl2_prod_id = "model1"

    level, rc = historylevel(datasequence_history_file, "DATA")
    assert level == 0
    assert rc == 0

    level, rc = historylevel(calibration_history_file, "PEDCALIB")
    assert level == 0
    assert rc == 0

    options.dl1_prod_id = "tailcut84"
    options.dl2_prod_id = "model2"

    level, rc = historylevel(datasequence_history_file, "DATA")
    assert level == 1
    assert rc == 0


def test_preparejobs(running_analysis_dir, sequence_list):
    from osa.job import prepare_jobs

    options.simulate = False
    options.directory = running_analysis_dir
    prepare_jobs(sequence_list)
    expected_calib_script = os.path.join(running_analysis_dir, "sequence_LST1_01809.py")
    expected_data_script = os.path.join(running_analysis_dir, "sequence_LST1_01807.py")
    assert os.path.isfile(os.path.abspath(expected_calib_script))
    assert os.path.isfile(os.path.abspath(expected_data_script))


def test_sequence_filenames(running_analysis_dir, sequence_list):
    from osa.job import sequence_filenames

    for sequence in sequence_list:
        sequence_filenames(sequence)
        assert sequence.script == running_analysis_dir / f"sequence_LST1_{sequence.run:05d}.py"


def test_scheduler_env_variables(sequence_list, running_analysis_dir):
    from osa.job import scheduler_env_variables

    # Extract the first sequence
    first_sequence = sequence_list[0]
    env_variables = scheduler_env_variables(first_sequence)
    assert env_variables == [
        "#SBATCH --job-name=LST1_01809",
        "#SBATCH --time=1:15:00",
        f"#SBATCH --chdir={running_analysis_dir}",
        "#SBATCH --output=log/Run01809.%4a_jobid_%A.out",
        "#SBATCH --error=log/Run01809.%4a_jobid_%A.err",
        f'#SBATCH --partition={cfg.get("SLURM", "PARTITION_PEDCALIB")}',
        "#SBATCH --mem-per-cpu=3GB",
    ]
    # Extract the second sequence
    second_sequence = sequence_list[1]
    env_variables = scheduler_env_variables(second_sequence)
    assert env_variables == [
        "#SBATCH --job-name=LST1_01807",
        "#SBATCH --time=1:15:00",
        f"#SBATCH --chdir={running_analysis_dir}",
        "#SBATCH --output=log/Run01807.%4a_jobid_%A.out",
        "#SBATCH --error=log/Run01807.%4a_jobid_%A.err",
        "#SBATCH --array=0-10",
        f'#SBATCH --partition={cfg.get("SLURM", "PARTITION_DATA")}',
        "#SBATCH --mem-per-cpu=16GB",
    ]


def test_job_header_template(sequence_list, running_analysis_dir):
    """Extract and check the header for the first two sequences."""
    from osa.job import job_header_template

    # Extract the first sequence
    first_sequence = sequence_list[0]
    header = job_header_template(first_sequence)
    output_string1 = dedent(
        f"""\
    #!/bin/env python

    #SBATCH --job-name=LST1_01809
    #SBATCH --time=1:15:00
    #SBATCH --chdir={running_analysis_dir}
    #SBATCH --output=log/Run01809.%4a_jobid_%A.out
    #SBATCH --error=log/Run01809.%4a_jobid_%A.err
    #SBATCH --partition={cfg.get('SLURM', 'PARTITION_PEDCALIB')}
    #SBATCH --mem-per-cpu=3GB"""
    )
    assert header == output_string1

    # Extract the second sequence
    second_sequence = sequence_list[1]
    header = job_header_template(second_sequence)
    output_string2 = dedent(
        f"""\
    #!/bin/env python

    #SBATCH --job-name=LST1_01807
    #SBATCH --time=1:15:00
    #SBATCH --chdir={running_analysis_dir}
    #SBATCH --output=log/Run01807.%4a_jobid_%A.out
    #SBATCH --error=log/Run01807.%4a_jobid_%A.err
    #SBATCH --array=0-10
    #SBATCH --partition={cfg.get('SLURM', 'PARTITION_DATA')}
    #SBATCH --mem-per-cpu=16GB"""
    )
    assert header == output_string2


def test_create_job_template_scheduler(
    sequence_list,
    drs4_time_calibration_files,
    drs4_baseline_file,
    calibration_file,
    run_summary_file,
    pedestal_ids_file,
):
    from osa.job import data_sequence_job_template

    assert pedestal_ids_file.exists()

    options.test = False
    options.simulate = False
    content1 = data_sequence_job_template(sequence_list[1])
    expected_content1 = dedent(
        f"""\
    #!/bin/env python

    #SBATCH --job-name=LST1_01807
    #SBATCH --time=1:15:00
    #SBATCH --chdir={Path.cwd()}/test_osa/test_files0/running_analysis/20200117/v0.1.0
    #SBATCH --output=log/Run01807.%4a_jobid_%A.out
    #SBATCH --error=log/Run01807.%4a_jobid_%A.err
    #SBATCH --array=0-10
    #SBATCH --partition={cfg.get('SLURM', 'PARTITION_DATA')}
    #SBATCH --mem-per-cpu={cfg.get('SLURM', 'MEMSIZE_DATA')}

    import os
    import subprocess
    import sys
    import tempfile

    os.environ['CTAPIPE_CACHE'] = '/fefs/aswg/lstanalyzer/.ctapipe/ctapipe_cache'
    os.environ['CTAPIPE_SVC_PATH'] = '/fefs/aswg/lstanalyzer/.ctapipe/service'
    os.environ['MPLCONFIGDIR'] = '/fefs/aswg/lstanalyzer/.cache/matplotlib'
    subruns = int(os.getenv('SLURM_ARRAY_TASK_ID'))

    with tempfile.TemporaryDirectory() as tmpdirname:
        os.environ['NUMBA_CACHE_DIR'] = tmpdirname
        proc = subprocess.run([
            'datasequence',
            '--config',
            '{DEFAULT_CFG}',
            '--date=2020-01-17',
            '--prod-id=v0.1.0',
            '--drs4-pedestal-file={drs4_baseline_file}',
            '--time-calib-file={drs4_time_calibration_files[0]}',
            '--pedcal-file={calibration_file}',
            '--systematic-correction-file={Path.cwd()}/test_osa/test_files0/monitoring/PixelCalibration/Cat-A/ffactor_systematics/20200725/pro/ffactor_systematics_20200725.h5',
            '--drive-file={Path.cwd()}/test_osa/test_files0/monitoring/DrivePositioning/DrivePosition_log_20200117.txt',
            '--run-summary={run_summary_file}',
            f'01807.{{subruns:04d}}',
            'LST1'
        ])

    sys.exit(proc.returncode)"""
    )

    content2 = data_sequence_job_template(sequence_list[2])
    expected_content2 = dedent(
        f"""\
        #!/bin/env python

        #SBATCH --job-name=LST1_01808
        #SBATCH --time=1:15:00
        #SBATCH --chdir={Path.cwd()}/test_osa/test_files0/running_analysis/20200117/v0.1.0
        #SBATCH --output=log/Run01808.%4a_jobid_%A.out
        #SBATCH --error=log/Run01808.%4a_jobid_%A.err
        #SBATCH --array=0-8
        #SBATCH --partition={cfg.get('SLURM', 'PARTITION_DATA')}
        #SBATCH --mem-per-cpu={cfg.get('SLURM', 'MEMSIZE_DATA')}

        import os
        import subprocess
        import sys
        import tempfile

        os.environ['CTAPIPE_CACHE'] = '/fefs/aswg/lstanalyzer/.ctapipe/ctapipe_cache'
        os.environ['CTAPIPE_SVC_PATH'] = '/fefs/aswg/lstanalyzer/.ctapipe/service'
        os.environ['MPLCONFIGDIR'] = '/fefs/aswg/lstanalyzer/.cache/matplotlib'
        subruns = int(os.getenv('SLURM_ARRAY_TASK_ID'))

        with tempfile.TemporaryDirectory() as tmpdirname:
            os.environ['NUMBA_CACHE_DIR'] = tmpdirname
            proc = subprocess.run([
                'datasequence',
                '--config',
                '{DEFAULT_CFG}',
                '--date=2020-01-17',
                '--prod-id=v0.1.0',
                '--drs4-pedestal-file={drs4_baseline_file}',
                '--time-calib-file={drs4_time_calibration_files[0]}',
                '--pedcal-file={calibration_file}',
                '--systematic-correction-file={Path.cwd()}/test_osa/test_files0/monitoring/PixelCalibration/Cat-A/ffactor_systematics/20200725/pro/ffactor_systematics_20200725.h5',
                '--drive-file={Path.cwd()}/test_osa/test_files0/monitoring/DrivePositioning/DrivePosition_log_20200117.txt',
                '--run-summary={run_summary_file}',
                f'--pedestal-ids-file={Path.cwd()}/test_osa/test_files0/auxiliary/PedestalFinder/20200117/pedestal_ids_Run01808.{{subruns:04d}}.h5',
                f'01808.{{subruns:04d}}',
                'LST1'
            ])

        sys.exit(proc.returncode)"""
    )

    options.simulate = True
    assert content1 == expected_content1
    assert content2 == expected_content2


def test_create_job_template_local(
    sequence_list,
    drs4_time_calibration_files,
    drs4_baseline_file,
    calibration_file,
    systematic_correction_files,
    run_summary_file,
    pedestal_ids_file,
    r0_data,
):
    """Check the job file in local mode (assuming no scheduler)."""
    from osa.job import data_sequence_job_template

    for file in drs4_time_calibration_files:
        assert file.exists()

    for file in systematic_correction_files:
        assert file.exists()

    for file in r0_data:
        assert file.exists()

    assert pedestal_ids_file.exists()

    options.test = True
    options.simulate = False

    content1 = data_sequence_job_template(sequence_list[1])
    expected_content1 = dedent(
        f"""\
    #!/bin/env python

    import os
    import subprocess
    import sys
    import tempfile

    subruns = 0

    with tempfile.TemporaryDirectory() as tmpdirname:
        os.environ['NUMBA_CACHE_DIR'] = tmpdirname
        proc = subprocess.run([
            'datasequence',
            '--config',
            '{DEFAULT_CFG}',
            '--date=2020-01-17',
            '--prod-id=v0.1.0',
            '--drs4-pedestal-file={drs4_baseline_file}',
            '--time-calib-file={drs4_time_calibration_files[0]}',
            '--pedcal-file={calibration_file}',
            '--systematic-correction-file={Path.cwd()}/test_osa/test_files0/monitoring/PixelCalibration/Cat-A/ffactor_systematics/20200725/pro/ffactor_systematics_20200725.h5',
            '--drive-file={Path.cwd()}/test_osa/test_files0/monitoring/DrivePositioning/DrivePosition_log_20200117.txt',
            '--run-summary={run_summary_file}',
            f'01807.{{subruns:04d}}',
            'LST1'
        ])

    sys.exit(proc.returncode)"""
    )

    content2 = data_sequence_job_template(sequence_list[2])
    expected_content2 = dedent(
        f"""\
        #!/bin/env python

        import os
        import subprocess
        import sys
        import tempfile

        subruns = 0

        with tempfile.TemporaryDirectory() as tmpdirname:
            os.environ['NUMBA_CACHE_DIR'] = tmpdirname
            proc = subprocess.run([
                'datasequence',
                '--config',
                '{DEFAULT_CFG}',
                '--date=2020-01-17',
                '--prod-id=v0.1.0',
                '--drs4-pedestal-file={drs4_baseline_file}',
                '--time-calib-file={drs4_time_calibration_files[0]}',
                '--pedcal-file={calibration_file}',
                '--systematic-correction-file={Path.cwd()}/test_osa/test_files0/monitoring/PixelCalibration/Cat-A/ffactor_systematics/20200725/pro/ffactor_systematics_20200725.h5',
                '--drive-file={Path.cwd()}/test_osa/test_files0/monitoring/DrivePositioning/DrivePosition_log_20200117.txt',
                '--run-summary={run_summary_file}',
                f'--pedestal-ids-file={Path.cwd()}/test_osa/test_files0/auxiliary/PedestalFinder/20200117/pedestal_ids_Run01808.{{subruns:04d}}.h5',
                f'01808.{{subruns:04d}}',
                'LST1'
            ])

        sys.exit(proc.returncode)"""
    )

    options.simulate = True

    assert content1 == expected_content1
    assert content2 == expected_content2


def test_create_job_scheduler_calibration(sequence_list):
    """Check the pilot job file for the calibration pipeline."""
    from osa.job import calibration_sequence_job_template

    options.test = True
    options.simulate = False
    content = calibration_sequence_job_template(sequence_list[0])
    expected_content = dedent(
        f"""\
    #!/bin/env python

    import os
    import subprocess
    import sys
    import tempfile

    subruns = 0

    with tempfile.TemporaryDirectory() as tmpdirname:
        os.environ['NUMBA_CACHE_DIR'] = tmpdirname
        proc = subprocess.run([
            'calibration_pipeline',
            '--config',
            '{DEFAULT_CFG}',
            '--date=2020-01-17',
            '--drs4-pedestal-run=01804',
            '--pedcal-run=01809',
            'LST1'
        ])

    sys.exit(proc.returncode)"""
    )
    options.simulate = True
    assert content == expected_content


def test_set_cache_dirs():
    from osa.job import set_cache_dirs

    cache = set_cache_dirs()
    cache_dirs = dedent(
        f"""\
    os.environ['CTAPIPE_CACHE'] = '{cfg.get('CACHE', 'CTAPIPE_CACHE')}'
    os.environ['CTAPIPE_SVC_PATH'] = '{cfg.get('CACHE', 'CTAPIPE_SVC_PATH')}'
    os.environ['MPLCONFIGDIR'] = '{cfg.get('CACHE', 'MPLCONFIGDIR')}'"""
    )
    assert cache_dirs == cache


def test_calibration_history_level():
    from osa.job import check_history_level

    levels = {"onsite_create_drs4_pedestal_file": 1, "onsite_create_calibration_file": 0}
    level, exit_status = check_history_level(calibration_history_file, levels)
    assert level == 0
    assert exit_status == 0


@pytest.fixture
def mock_sacct_output():
    """Mock output of sacct to be able to use it in get_squeue_output function."""
    return Path("./extra") / "sacct_output.csv"


@pytest.fixture
def mock_squeue_output():
    """Mock output of squeue to be able to use it in get_squeue_output function."""
    return Path("./extra") / "squeue_output.csv"


@pytest.fixture
def sacct_output(mock_sacct_output):
    from osa.job import get_sacct_output

    return get_sacct_output(mock_sacct_output)


@pytest.fixture
def squeue_output(mock_squeue_output):
    from osa.job import get_squeue_output

    return get_squeue_output(mock_squeue_output)


def test_set_queue_values(sacct_output, squeue_output, sequence_list):
    from osa.job import set_queue_values

    set_queue_values(
        sacct_info=sacct_output,
        squeue_info=squeue_output,
        sequence_list=sequence_list,
    )
    # Running calibration sequence
    assert sequence_list[0].state == "RUNNING"
    assert sequence_list[0].exit is None
    assert sequence_list[0].jobid == 12951086
    assert sequence_list[0].cputime == "00:36:00"
    assert sequence_list[0].tries == 4
    # Pending DATA sequences
    assert sequence_list[1].state == "PENDING"
    assert sequence_list[1].tries == 2
    assert sequence_list[1].exit is None
    assert sequence_list[2].state == "PENDING"
    assert sequence_list[2].exit is None
    assert sequence_list[2].tries == 1


def test_plot_job_statistics(sacct_output, running_analysis_dir):
    from osa.job import plot_job_statistics

    log_dir = running_analysis_dir / "log"
    log_dir.mkdir(parents=True, exist_ok=True)
    assert log_dir.exists()
    plot_job_statistics(sacct_output, log_dir)
    plot_file = log_dir / "job_statistics.pdf"
    assert plot_file.exists()



"""Script to reprocess all daily longterm files found under a given production."""

import logging
import subprocess as sp
from pathlib import Path

import click

from osa.utils.logging import myLogger

ANALYSIS_PATH = Path("/fefs/aswg/data/real")
LONGTERM_PATH = ANALYSIS_PATH / "OSA/DL1DataCheck_LongTerm"


log = myLogger(logging.getLogger())


def run_longterm(date: str, prod_id: str, new_prod_id: str, log_dir: Path):
    """
    Run the longterm script for a given date.

    Parameters
    ----------
    date : str
        Date to reprocess in YYYYMMDD format.
    prod_id : str
        Production ID to reprocess.
    new_prod_id : str
        New production ID to reprocess.
    log_dir : Path
        Path to the directory where job logs will be stored.
    """
    dl1_dir = ANALYSIS_PATH / "DL1" / date / prod_id / "tailcut84" / "datacheck"
    muons_dir = ANALYSIS_PATH / "DL1" / date / prod_id / "muons"
    new_longterm_dir = LONGTERM_PATH / new_prod_id / date
    longterm_output_file = new_longterm_dir / f"DL1_datacheck_{date}.h5"

    log_file = log_dir / f"daily_check_{date}_%j.log"

    cmd = [
        "sbatch",
        "-o",
        str(log_file),
        "lstchain_longterm_dl1_check",
        f"--input-dir={dl1_dir}",
        f"--output-file={longterm_output_file}",
        f"--muons-dir={muons_dir}",
        "--batch",
    ]

    sp.run(cmd)
    log.info(f"Executing {' '.join(cmd)}")


@click.command()
@click.option("--prod-id", help="Production ID to reprocess")
@click.option("--new-prod-id", help="New production ID")
@click.option("--log-dir", type=click.Path(path_type=Path), help="Path to log directory")
def main(prod_id: str = None, new_prod_id: str = None, log_dir: Path = None):
    """
    Reprocess the daily check files for all existing dates found on a given production.
    """
    log_dir.mkdir(exist_ok=True, parents=True)
    longterm_dir = LONGTERM_PATH / prod_id

    list_of_processed_dates = longterm_dir.glob("20*")

    for date_directory in list_of_processed_dates:
        run_longterm(date_directory.name, prod_id, new_prod_id, log_dir)


if __name__ == "__main__":
    main()



"""Script to update the source coordinates in the RunCatalog ECSV files."""

import logging
from pathlib import Path

import click
import numpy as np
from astropy.coordinates import SkyCoord, name_resolve
from astropy.table import Table

from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))


@click.command()
@click.argument("file", type=click.Path(exists=True))
@click.option("--source", type=str)
@click.option("--ra", type=float, help="Right Ascension in degrees")
@click.option("--dec", type=float, help="Declination in degrees")
def main(file: Path = None, source: str = None, ra: float = np.nan, dec: float = np.nan):
    """Update the source coordinates in the RunCatalog ECSV files."""
    log.setLevel(logging.INFO)

    table = Table.read(file)
    table.add_index("run_id")

    for run in table:
        source_name = run["source_name"]
        try:
            coords = SkyCoord.from_name(source_name)
            run["source_ra"] = coords.ra.deg
            run["source_dec"] = coords.dec.deg

        except name_resolve.NameResolveError:
            if source is None:
                log.warning(
                    f"Could not resolve coordinates for {source_name}. "
                    "Add coordinates through the command line."
                )
            if source_name == source:
                run["source_ra"] = ra
                run["source_dec"] = dec

    log.info(f"Updated coordinates in {file}")
    table.write(file, format="ascii.ecsv", overwrite=True, delimiter=",")


if __name__ == "__main__":
    main()



from pathlib import Path


def test_failed_history():
    from osa.veto import failed_history

    good_history_file = Path("./extra/history_files/sequence_LST1_04183.history")
    bad_history_file = Path("./extra/history_files/sequence_LST1_04183_failed.history")
    one_line_history_file = Path("./extra/history_files/sequence_LST1_04183_oneline.history")
    one_line_history = failed_history(one_line_history_file)
    good_history_failed = failed_history(good_history_file)
    bad_history_failed = failed_history(bad_history_file)

    assert good_history_failed is False
    assert one_line_history is False
    assert bad_history_failed is True


def test_get_veto_list(sequence_list):
    from osa.veto import get_veto_list

    veto_list = get_veto_list(sequence_list)
    assert not veto_list


def test_get_closed_list(sequence_list):
    from osa.veto import get_closed_list

    closed_list = get_closed_list(sequence_list)
    seq_list = ["LST1_01807", "LST1_01808", "LST1_01809"]
    for sequence in seq_list:
        assert sequence in closed_list



"""Functionality for sending emails."""

import smtplib
from email.message import EmailMessage
from osa.configs.config import cfg


__all__ = ["send_warning_mail"]


def send_warning_mail(date: str):
    """Send a warning email to the admin."""
    msg = EmailMessage()
    msg["Subject"] = f"WARNING: {date} is not closed yet"
    msg["From"] = "lstanalyzer"
    msg["To"] = cfg.get("mail", "recipient")
    msg.set_content(f"Could not close LST1 for {date}. Please check.")

    with smtplib.SMTP("localhost") as email:
        email.send_message(msg)



from datetime import datetime
from pathlib import Path

from osa.configs.config import cfg


def test_get_night_summary_file():
    from osa.nightsummary.nightsummary import get_run_summary_file

    date = datetime.fromisoformat("2020-01-01")
    summary_filename = get_run_summary_file(date)
    assert summary_filename == Path(cfg.get("LST1", "RUN_SUMMARY_DIR")) / "RunSummary_20200101.ecsv"


def test_run_summary_table(run_summary_file):
    from osa.nightsummary.nightsummary import run_summary_table

    assert run_summary_file.exists()

    date = datetime.fromisoformat("2020-01-17")
    summary = run_summary_table(date)
    assert "run_id" in summary.columns


def test_produce_run_summary_file():
    from osa.nightsummary.nightsummary import produce_run_summary_file

    date = datetime.fromisoformat("2020-01-01")
    produce_run_summary_file(date)
    summary_file = Path(cfg.get("LST1", "RUN_SUMMARY_DIR")) / "RunSummary_20200101.ecsv"
    assert summary_file.exists()



from osa.utils.iofile import append_to_file
from osa.configs import options


def test_write_and_append_to_file(txt_file_test):
    assert txt_file_test.exists()
    assert txt_file_test.read_text() == "This is a test"
    options.simulate = False
    append_to_file(txt_file_test, "\nAnother line")
    options.simulate = True
    assert txt_file_test.read_text() == "This is a test\nAnother line"



"""Script called from the batch scheduler to process a run."""

import logging
import sys
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.job import historylevel
from osa.workflow.stages import AnalysisStage
from osa.provenance.capture import trace
from osa.utils.cliopts import data_sequence_cli_parsing
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_dir

__all__ = ["data_sequence", "r0_to_dl1", "dl1_to_dl2", "dl1ab", "dl1_datacheck"]

log = myLogger(logging.getLogger())


def data_sequence(
    calibration_file: Path,
    pedestal_file: Path,
    time_calibration_file: Path,
    systematic_correction_file: Path,
    drive_file: Path,
    run_summary: Path,
    pedestal_ids_file: Path,
    run_str: str,
):
    """
    Performs all the steps to process a whole run.

    Parameters
    ----------
    calibration_file: pathlib.Path
    pedestal_file: pathlib.Path
    time_calibration_file: pathlib.Path
    systematic_correction_file: pathlib.Path
    drive_file: pathlib.Path
    run_summary: pathlib.Path
    pedestal_ids_file: pathlib.Path
    run_str: str

    Returns
    -------
    rc: int
        Return code of the last executed command.
    """
    history_file = Path(options.directory) / f"sequence_{options.tel_id}_{run_str}.history"
    # Set the starting level and corresponding return code from last analysis step
    # registered in the history file.
    level, rc = (4, 0) if options.simulate else historylevel(history_file, "DATA")
    log.info(f"Going to level {level}")

    if level == 4:
        rc = r0_to_dl1(
            calibration_file,
            pedestal_file,
            time_calibration_file,
            systematic_correction_file,
            drive_file,
            run_summary,
            pedestal_ids_file,
            run_str,
        )
        level -= 1
        log.info(f"Going to level {level}")

    if level == 3:
        rc = dl1ab(run_str)
        if cfg.getboolean("lstchain", "store_image_dl1ab"):
            level -= 1
            log.info(f"Going to level {level}")
        else:
            level -= 2
            log.info(f"No images stored in dl1ab. Producing DL2. Going to level {level}")

    if level == 2:
        rc = dl1_datacheck(run_str)
        if options.no_dl2:
            level = 0
            log.info(f"No DL2 are going to be produced. Going to level {level}")
        else:
            level -= 1
            log.info(f"Going to level {level}")

    if level == 1:
        if options.no_dl2:
            level = 0
            log.info(f"No DL2 are going to be produced. Going to level {level}")
        else:
            rc = dl1_to_dl2(run_str)
            level -= 1
            log.info(f"Going to level {level}")

    if level == 0:
        log.info(f"Job for sequence {run_str} finished without fatal errors")

    return rc


@trace
def r0_to_dl1(
    calibration_file: Path,
    pedestal_file: Path,
    time_calibration_file: Path,
    systematic_correction_file: Path,
    drive_file: Path,
    run_summary: Path,
    pedestal_ids_file: Path,
    run_str: str,
) -> int:
    """
    Prepare and launch the actual lstchain script that is performing
    the low and high-level calibration to raw camera images.
    It also applies the image cleaning and obtains shower parameters.

    Parameters
    ----------
    calibration_file: pathlib.Path
    pedestal_file: pathlib.Path
    time_calibration_file: pathlib.Path
    systematic_correction_file: pathlib.Path
    drive_file: pathlib.Path
    run_summary: : pathlib.Path
        Path to the run summary file
    pedestal_ids_file: pathlib.Path
        Path to file containing the interleaved pedestal event ids
    run_str: str
        XXXXX.XXXX (run_number.subrun_number)

    Returns
    -------
    rc: int
        Return code of the executed command.
    """
    command = cfg.get("lstchain", "r0_to_dl1")
    night_dir = date_to_dir(options.date)
    r0_dir = Path(cfg.get("LST1", "R0_DIR")) / night_dir
    r0_file = r0_dir / f"LST-1.1.Run{run_str}.fits.fz"
    dl1a_config = Path(cfg.get("lstchain", "dl1a_config"))

    cmd = [
        command,
        f"--input-file={r0_file}",
        f"--output-dir={options.directory}",
        f"--pedestal-file={pedestal_file}",
        f"--calibration-file={calibration_file}",
        f"--time-calibration-file={time_calibration_file}",
        f"--systematic-correction-file={systematic_correction_file}",
        f"--config={dl1a_config}",
        f"--pointing-file={drive_file}",
        f"--run-summary-path={run_summary}",
    ]

    if pedestal_ids_file is not None:
        cmd.append(f"--pedestal-ids-path={pedestal_ids_file}")

    if options.simulate:
        return 0

    analysis_step = AnalysisStage(run=run_str, command_args=cmd, config_file=dl1a_config.name)
    analysis_step.execute()
    return analysis_step.rc


@trace
def dl1ab(run_str: str) -> int:
    """
    Prepare and launch the actual lstchain script that is performing
    the image cleaning considering the interleaved pedestal information
    and obtains shower parameters. It keeps the shower images.

    Parameters
    ----------
    run_str: str

    Returns
    -------
    rc: int
        Return code of the executed command.
    """
    # Create a new subdirectory for the dl1ab output
    dl1ab_subdirectory = Path(options.directory) / options.dl1_prod_id
    dl1ab_subdirectory.mkdir(parents=True, exist_ok=True)
    dl1b_config = Path(cfg.get("lstchain", "dl1b_config"))
    # DL1a input file from base running_analysis directory
    input_dl1_datafile = Path(options.directory) / f"dl1_LST-1.Run{run_str}.h5"
    # DL1b output file to be stored in the dl1ab subdirectory
    output_dl1_datafile = dl1ab_subdirectory / f"dl1_LST-1.Run{run_str}.h5"

    # Prepare and launch the actual lstchain script
    command = cfg.get("lstchain", "dl1ab")
    cmd = [
        command,
        f"--input-file={input_dl1_datafile}",
        f"--output-file={output_dl1_datafile}",
        f"--config={dl1b_config}",
    ]
    if not cfg.getboolean("lstchain", "store_image_dl1ab"):
        cmd.append("--no-image=True")

    if options.simulate:
        return 0

    analysis_step = AnalysisStage(run=run_str, command_args=cmd, config_file=dl1b_config.name)
    analysis_step.execute()
    return analysis_step.rc


@trace
def dl1_datacheck(run_str: str) -> int:
    """
    Run datacheck script

    Parameters
    ----------
    run_str: str

    Returns
    -------
    rc: int
    """
    # Create a new subdirectory for the dl1ab output
    dl1ab_subdirectory = Path(options.directory) / options.dl1_prod_id
    input_dl1_datafile = dl1ab_subdirectory / f"dl1_LST-1.Run{run_str}.h5"
    output_directory = Path(options.directory) / options.dl1_prod_id
    output_directory.mkdir(parents=True, exist_ok=True)

    # Prepare and launch the actual lstchain script
    command = cfg.get("lstchain", "check_dl1")
    cmd = [
        command,
        f"--input-file={input_dl1_datafile}",
        f"--output-dir={output_directory}",
        f"--muons-dir={options.directory}",
        "--omit-pdf",
        "--batch",
    ]

    if options.simulate:
        return 0

    analysis_step = AnalysisStage(run=run_str, command_args=cmd)
    analysis_step.execute()
    return analysis_step.rc


@trace
def dl1_to_dl2(run_str: str) -> int:
    """
    It prepares and execute the dl1 to dl2 lstchain scripts that applies
    the already trained RFs models to DL1 files. It identifies the
    primary particle, reconstructs its energy and direction.

    Parameters
    ----------
    run_str: str

    Returns
    -------
    rc: int
    """
    dl1ab_subdirectory = Path(options.directory) / options.dl1_prod_id
    dl2_subdirectory = Path(options.directory) / options.dl2_prod_id
    dl2_config = Path(cfg.get("lstchain", "dl2_config"))
    rf_models_directory = Path(cfg.get("lstchain", "RF_MODELS"))
    dl1_file = dl1ab_subdirectory / f"dl1_LST-1.Run{run_str}.h5"

    command = cfg.get("lstchain", "dl1_to_dl2")
    cmd = [
        command,
        f"--input-file={dl1_file}",
        f"--output-dir={dl2_subdirectory}",
        f"--path-models={rf_models_directory}",
        f"--config={dl2_config}",
    ]

    if options.simulate:
        return 0

    analysis_step = AnalysisStage(run=run_str, command_args=cmd, config_file=dl2_config.name)
    analysis_step.execute()
    return analysis_step.rc


def main():
    """Performs the analysis steps to convert raw data into DL2 files."""
    (
        calibration_file,
        drs4_ped_file,
        time_calibration_file,
        systematic_correction_file,
        drive_log_file,
        run_summary_file,
        pedestal_ids_file,
        run_number,
    ) = data_sequence_cli_parsing()

    if options.verbose:
        log.setLevel(logging.DEBUG)
    else:
        log.setLevel(logging.INFO)

    # Run the routine piping all the analysis steps
    rc = data_sequence(
        calibration_file,
        drs4_ped_file,
        time_calibration_file,
        systematic_correction_file,
        drive_log_file,
        run_summary_file,
        pedestal_ids_file,
        run_number,
    )
    sys.exit(rc)


if __name__ == "__main__":
    main()



from datetime import datetime
from osa.configs import options


def test_finished_assignments(sequence_list):
    from osa.report import finished_assignments

    finished_dict = finished_assignments(sequence_list)
    now = datetime.utcnow().strftime("%Y-%m-%d %H:%M")

    assert finished_dict["END"] == now
    assert finished_dict["NIGHT"] == "2020-01-17"
    assert finished_dict["TELESCOPE"] == "LST1"
    assert finished_dict["RAW_GB"] == 0
    assert finished_dict["FILES_RAW"] == 25
    assert finished_dict["SEQUENCES"] == 3


def test_history(base_test_dir):
    from osa.report import history

    run = "01800"
    prod_id = "v1.0.0"
    program = "r0_to_dl1"
    input_file = "r0_to_dl1_01800.fits"
    input_card = "r0_dl1.config"
    rc = 0
    history_file = base_test_dir / "r0_to_dl1_01800.history"
    date_string = datetime.utcnow().isoformat(sep=" ", timespec="minutes")

    options.simulate = False

    history(
        run=run,
        prod_id=prod_id,
        stage=program,
        return_code=rc,
        history_file=history_file,
        input_file=input_file,
        config_file=input_card,
    )

    options.simulate = True

    logged_string = (
        f"01800 r0_to_dl1 v1.0.0 {date_string} " "r0_to_dl1_01800.fits r0_dl1.config 0\n"
    )

    assert history_file.exists()
    assert history_file.read_text() == logged_string



"""Functions to deal with dates and prod IDs."""


import inspect
import logging
import os
import time
from datetime import datetime, timedelta
from pathlib import Path
from socket import gethostname

import osa.paths
from osa.configs import options
from osa.configs.config import cfg
from osa.utils.iofile import write_to_file
from osa.utils.logging import myLogger

__all__ = [
    "get_lstchain_version",
    "date_to_dir",
    "date_to_iso",
    "is_day_closed",
    "get_prod_id",
    "night_finished_flag",
    "is_defined",
    "create_lock",
    "stringify",
    "gettag",
    "get_dl1_prod_id",
    "get_dl2_prod_id",
    "time_to_seconds",
    "DATACHECK_FILE_PATTERNS",
    "YESTERDAY",
    "set_prod_ids",
    "is_night_time",
    "cron_lock",
    "example_seq",
    "wait_for_daytime",
]

log = myLogger(logging.getLogger(__name__))

DATACHECK_PRODUCTS = ["drs4", "enf_calibration", "dl1"]

DATACHECK_FILE_PATTERNS = {
    "PEDESTAL": "drs4*.pdf",
    "CALIB": "calibration*.pdf",
    "DL1AB": "datacheck_dl1*.pdf",
    "LONGTERM": "DL1_datacheck_*.*",
}

YESTERDAY = datetime.now() - timedelta(days=1)


def get_lstchain_version():
    """
    Get the lstchain version.

    Returns
    -------
    lstchain_version: string
    """
    from lstchain import __version__

    return f"v{__version__}"


def get_prod_id():
    """
    Get production ID from the configuration file if it is defined.
    Otherwise, it takes the lstchain version used.

    Returns
    -------
    prod_id: string
    """
    if not options.prod_id:
        if cfg.get("LST1", "PROD_ID") is not None:
            options.prod_id = cfg.get("LST1", "PROD_ID")
        else:
            options.prod_id = get_lstchain_version()

    log.debug(f"Getting prod ID for the running analysis directory: {options.prod_id}")

    return options.prod_id


def get_dl1_prod_id():
    """
    Get the prod ID for the dl1 products provided
    it is defined in the configuration file.

    Returns
    -------
    dl1_prod_id: string
    """
    if not options.dl1_prod_id:
        if cfg.get("LST1", "DL1_PROD_ID") is not None:
            options.dl1_prod_id = cfg.get("LST1", "DL1_PROD_ID")
        else:
            options.dl1_prod_id = get_lstchain_version()

    log.debug(f"Getting prod ID for DL1 products: {options.dl1_prod_id}")

    return options.dl1_prod_id


def get_dl2_prod_id():
    """

    Returns
    -------

    """
    if not options.dl2_prod_id:
        if cfg.get("LST1", "DL2_PROD_ID") is not None:
            options.dl2_prod_id = cfg.get("LST1", "DL2_PROD_ID")
        else:
            options.dl2_prod_id = get_lstchain_version()

    log.debug(f"Getting prod ID for DL2 products: {options.dl2_prod_id}")

    return options.dl2_prod_id


def create_lock(lockfile) -> bool:
    """
    Create a lock file to prevent multiple instances of the same analysis.

    Parameters
    ----------
    lockfile: pathlib.Path

    Returns
    -------
    bool
    """
    directory_lock = lockfile.parent
    if options.simulate:
        log.debug(f"Simulate the creation of lock file {lockfile}")
    elif lockfile.exists() and lockfile.is_file():
        with open(lockfile, "r") as f:
            hostpid = f.readline()
        log.error(f"Lock by a previous process {hostpid}, exiting!\n")
        return True
    else:
        if not directory_lock.exists():
            directory_lock.mkdir(exist_ok=True, parents=True)
            log.debug(f"Creating parent directory {directory_lock} for lock file")
            pid = str(os.getpid())
            hostname = gethostname()
            content = f"{hostname}:{pid}"
            write_to_file(lockfile, content)
            log.debug(f"Lock file {lockfile} created")
            return True
    return False


def night_finished_flag() -> Path:
    """
    Create night-is-finished lock file.

    Returns
    -------
    lockfile: pathlib.Path
        Path of the lock file
    """
    basename = cfg.get("LSTOSA", "end_of_activity")
    date = date_to_dir(options.date)
    close_directory = Path(cfg.get(options.tel_id, "CLOSER_DIR"))
    lock_file = close_directory / date / options.prod_id / basename
    log.debug(f"Looking for lock file {lock_file}")
    return lock_file.resolve()


def date_to_iso(date: datetime) -> str:
    """Function to change from YYYY-MM-DD to YYYY-MM-DD."""
    return date.strftime("%Y-%m-%d")


def date_to_dir(date: datetime) -> str:
    """Function to change from YYYY-MM-DD to YYYYMMDD format (used for directories)."""
    return date.strftime("%Y%m%d")


def is_defined(variable):
    """Check if a variable is already defined."""
    try:
        variable
    except NameError:
        variable = None
    return variable is not None


def is_day_closed() -> bool:
    """Get the name and Check for the existence of the Closer flag file."""
    flag_file = night_finished_flag()
    return flag_file.exists()


def stringify(args):
    """Join a list of arguments in a string."""
    return " ".join(map(str, args))


def gettag():
    """Get the name of the script currently being used."""
    parent_file = os.path.basename(inspect.stack()[1][1])
    parent_module = inspect.stack()[1][3]
    return f"{parent_file}({parent_module})"


def time_to_seconds(timestring):
    """
    Transform (D-)HH:MM:SS time format to seconds.

    Parameters
    ----------
    timestring: str or None
        Time in format (D-)HH:MM:SS

    Returns
    -------
    Seconds that correspond to (D-)HH:MM:SS
    """
    if timestring is None:
        timestring = "00:00:00"
    if "-" in timestring:
        # Day is also specified (D-)HH:MM:SS
        days, hhmmss = timestring.split("-")
        hours, minutes, seconds = hhmmss.split(":")
        return int(days) * 24 * 3600 + int(hours) * 3600 + int(minutes) * 60 + int(seconds)

    split_time = timestring.split(":")
    if len(split_time) == 2:
        # MM:SS
        minutes, seconds = split_time
        hours = 0
    elif len(split_time) == 3:
        # HH:MM:SS
        hours, minutes, seconds = split_time
    else:
        raise ValueError("Time format not recognized.")
    return int(hours) * 3600 + int(minutes) * 60 + int(seconds)


def set_prod_ids():
    """Set the product IDs."""
    options.prod_id = get_prod_id()

    if cfg.get("LST1", "DL1_PROD_ID") is not None:
        options.dl1_prod_id = get_dl1_prod_id()
    else:
        options.dl1_prod_id = options.prod_id

    if cfg.get("LST1", "DL2_PROD_ID") is not None:
        options.dl2_prod_id = get_dl2_prod_id()
    else:
        options.dl2_prod_id = options.prod_id


def is_night_time(hour):
    """Check if it is nighttime."""
    if 8 <= hour <= 18:
        return False
    log.error("It is dark outside...")
    return True


def example_seq():
    """Example sequence table output for testing."""
    return "./extra/example_sequencer.txt"


def cron_lock(tel) -> Path:
    """Create a lock file for the cron jobs."""
    return osa.paths.analysis_path(tel) / "cron.lock"


def wait_for_daytime(start=8, end=18):
    """
    Check every hour if it is still nighttime
    to not running jobs while it is still night.
    """
    while time.localtime().tm_hour <= start or time.localtime().tm_hour >= end:
        log.info("Waiting for sunrise to not interfere with the data-taking. Sleeping.")
        time.sleep(3600)



import pytest


@pytest.mark.tryfirst
def test_basedir_available(
    base_test_dir, r0_data, drs4_time_calibration_files, calibration_file, drs4_baseline_file
):

    assert base_test_dir.is_dir()

    for data_file in r0_data:
        assert data_file.exists()

    for file in drs4_time_calibration_files:
        assert file.exists()

    assert drs4_baseline_file.exists()
    assert calibration_file.exists()



"""
Script to handle the automatic closing of the OSA
checking that all jobs are correctly finished.
"""

import datetime
import glob
import logging
import os
import subprocess
import sys
from pathlib import Path

from osa.configs import options
from osa.paths import analysis_path
from osa.utils.cliopts import autocloser_cli_parser
from osa.utils.logging import myLogger
from osa.utils.mail import send_warning_mail
from osa.utils.utils import (
    night_finished_flag,
    is_day_closed,
    get_prod_id,
    is_night_time,
    cron_lock,
    example_seq,
    date_to_iso,
)

__all__ = ["Telescope", "Sequence"]

log = myLogger(logging.getLogger())


class Telescope:
    """Handle the telescope sequences, simulate and check them."""

    def __init__(
        self,
        telescope,
        date,
        config_file: Path,
        ignore_cronlock: bool = False,
        test: bool = False,
    ):
        """
        Parameters
        ----------
        telescope : str
            Options: LST1
        date : str
            Date in format YYYY-MM-DD
        config_file : pathlib.Path
            Path to the configuration file
        ignore_cronlock : bool
            Ignore cron lock file
        test : bool
            Run sequencer in test mode
        """

        self.telescope = telescope
        # necessary to make sure that cron.lock gets deleted in the end
        self.cron_lock = cron_lock(self.telescope)
        self.keyLine = None
        self.locked = False
        self.closed = False
        self.header_lines = []
        self.data_lines = []
        self.sequences = []
        self.seq_lines = None
        self.stdout = None
        self.stderr = None

        if self.is_closed():
            log.info(f"{self.telescope} is already closed! Ignoring {self.telescope}")
            return
        if not analysis_path(self.telescope).exists():
            log.warning(
                f"Analysis directory does not exist for {self.telescope}! "
                f"Ignoring {self.telescope}"
            )
            return
        if not self.lock_automatic_sequencer() and not ignore_cronlock:
            log.warning(f"{self.telescope} already locked! Ignoring {self.telescope}")
            return
        if not self.simulate_sequencer(date, config_file, test):
            log.warning(
                f"Simulation of the sequencer failed "
                f"for {self.telescope}! Ignoring {self.telescope}"
            )
            return

        self.parse_sequencer()

        if not self.build_sequences():
            log.info(f"Sequencer for {self.telescope} is empty! Exiting.")
            sys.exit()

    def __iter__(self):
        return iter(self.sequences)

    def __del__(self):
        """Delete cron lock file if it exists."""
        if self.locked:
            log.debug(f"Deleting {self.cron_lock}")
            os.remove(self.cron_lock)

    def is_closed(self):
        """Check if night is finished flag exists."""
        log.debug(f"Checking if {self.telescope} is closed")
        if night_finished_flag().exists():
            self.closed = True
            return True
        return False

    def lock_automatic_sequencer(self):
        """Check for cron lock file or create it if it does not exist."""
        if cron_lock(self.telescope).exists():
            return False
        log.debug(f"Creating {cron_lock(self.telescope)}")
        cron_lock(self.telescope).touch()
        self.locked = True
        return True

    def simulate_sequencer(self, date: str, config_file: Path, test: bool):
        """Launch the sequencer in simulation mode."""
        if test:
            self.read_file()
        else:
            sequencer_cmd = [
                "sequencer",
                "-s",
                "-c",
                str(config_file),
                "-d",
                date,
                self.telescope,
            ]
            log.debug(f"Executing {' '.join(sequencer_cmd)}")
            sequencer = subprocess.Popen(
                sequencer_cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
            self.stdout, self.stderr = sequencer.communicate()
            log.info(self.stdout)
            if sequencer.returncode != 0:
                log.warning(f"Sequencer returns error code {sequencer.returncode}")
                return False
            self.seq_lines = self.stdout.split("\n")
        return True

    def read_file(self):
        """Read an example sequencer output."""
        log.debug(f"Reading example of a sequencer output {example_seq()}")
        with open(example_seq(), "r") as self.stdout:
            stdout_tmp = self.stdout.read()
            log.info(stdout_tmp)
            self.seq_lines = stdout_tmp.split("\n")

    def parse_sequencer(self):
        """Parse the sequencer output lines."""
        log.debug(f"Parsing sequencer table of {self.telescope}")
        header = True
        data = False
        for line in self.seq_lines:
            if data and line:
                if line.startswith("LST1"):
                    self.data_lines.append(line)
            elif "Tel   Seq" in line:
                data = True
                header = False
                self.keyLine = line
            elif header:
                self.header_lines.append(line)

    def build_sequences(self):
        """Build the sequences and return True if there are any."""
        log.debug(f"Creating Sequence objects for {self.telescope}")
        self.sequences = [Sequence(self.keyLine, line) for line in self.data_lines]
        return bool(self.sequences)

    def close(
        self,
        date: str,
        config: Path,
        no_dl2: bool,
        simulate: bool = False,
        test: bool = False,
    ):
        """Launch the closer command."""
        log.info("Closing...")

        closer_cmd = [
            "closer",
            "-c",
            str(config),
            "-y",
            "-d",
            date,
            self.telescope,
        ]

        if simulate:
            closer_cmd.insert(1, "-s")

        if no_dl2:
            closer_cmd.insert(1, "--no-dl2")

        if test:
            self.closed = True
            return True

        log.debug(f"Executing {' '.join(closer_cmd)}")
        closer = subprocess.Popen(
            closer_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False
        )
        stdout, _ = closer.communicate()
        if closer.returncode != 0:
            log.warning(f"closer returned error code {closer.returncode}! See output: {stdout}")
            return False
        self.closed = True
        return True


class Sequence:
    """
    As for now the keys for the 'dict_sequence' are:
    (LST1) Tel Seq Parent Type Run Subruns Source Wobble Action Tries JobID
    State CPU_time Exit DL1% MUONS% DL1AB% DATACHECK% DL2%

    All the values in the 'dict_sequence' are strings
    """

    def __init__(self, keyLine, sequence):
        self.dict_sequence = {}
        self.keyLine = keyLine
        self.sequence = sequence
        self.understood = False
        self.readyToClose = False
        self.discarded = False
        self.closed = False

        self.parse_sequence()

    def parse_sequence(self):
        log.debug("Parsing sequence")
        self.dict_sequence = dict(zip(self.keyLine.split(), self.sequence.split()))
        log.debug(self.dict_sequence)

    def is_closed(self):
        return self.dict_sequence["Action"] == "Closed"

    def is_running(self):
        return self.dict_sequence["State"] == "RUNNING"

    def is_complete(self):
        return self.dict_sequence["State"] == "COMPLETED"

    def is_on_hold(self):
        return self.dict_sequence["State"] == "PENDING"

    def is_100(self, no_dl2: bool):
        """Check that all analysis products are 100% complete."""
        if (
            no_dl2
            and self.dict_sequence["Tel"] != "ST"
            and self.dict_sequence["DL1%"] == "100"
            and self.dict_sequence["DL1AB%"] == "100"
            and self.dict_sequence["MUONS%"] == "100"
        ):
            return True

        if (
            self.dict_sequence["Tel"] != "ST"
            and self.dict_sequence["DL1%"] == "100"
            and self.dict_sequence["DL1AB%"] == "100"
            and self.dict_sequence["MUONS%"] == "100"
            and self.dict_sequence["DL2%"] == "100"
        ):
            return True

        return False

    def is_flawless(self, no_dl2: bool):
        """Check that all jobs statuses are completed."""
        log.debug("Check if flawless")
        if (
            self.dict_sequence["Type"] == "DATA"
            and self.dict_sequence["Exit"] == "0:0"
            and self.is_100(no_dl2=no_dl2)
            and self.dict_sequence["State"] == "COMPLETED"
        ):
            return True
        if (
            self.dict_sequence["Type"] == "PEDCALIB"
            and self.dict_sequence["Exit"] == "0:0"
            and self.dict_sequence["State"] == "COMPLETED"
        ):
            return True

        return False

    def has_all_subruns(self):
        """
        Check that all subruns are complete by checking the
        total number of subrun wise DL1 files.
        """
        if self.dict_sequence["Type"] == "PEDCALIB":
            log.debug("Cannot check for missing subruns for CALIBRATION sequence")
            return True
        search_str = (
            f"{analysis_path(self.dict_sequence['Tel'])}/" f"dl1*{self.dict_sequence['Run']}*.h5"
        )
        subrun_nrs = sorted(
            [int(os.path.basename(file).split(".")[2]) for file in glob.glob(search_str)]
        )
        return bool(subrun_nrs and len(subrun_nrs) == int(self.dict_sequence["Subruns"]))

    def close(
        self, date: str, config: Path, no_dl2: bool, simulate: bool = False, test: bool = False
    ):
        """Close the sequence by calling the 'closer' script for a given sequence."""
        log.info("Closing sequence...")

        closer_cmd = [
            "closer",
            "-c",
            str(config),
            "-y",
            "-d",
            date,
            f"--seq={self.dict_sequence['Run']}",
            self.dict_sequence["Tel"],
        ]

        if simulate:
            closer_cmd.insert(1, "-s")

        if no_dl2:
            closer_cmd.insert(1, "--no-dl2")

        if test:
            self.closed = True
            return True

        log.debug(f"Executing {' '.join(closer_cmd)}")
        closer = subprocess.Popen(closer_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        stdout, _ = closer.communicate()
        if closer.returncode != 0:
            log.warning(f"closer returned error code {closer.returncode}! See output: {stdout}")
            return False

        self.closed = True

        return True


def understand_sequence(seq, no_dl2: bool):
    """Check if sequence is completed and ready to be closed."""
    if seq.is_closed():
        seq.understood = True
        log.info("Is closed")
        seq.closed = True
        seq.readyToClose = True
        return True

    if not seq.is_complete():
        log.info("Is not completed yet")
        return True

    #  returns True if check is not possible, e.i. for ST and CALIBRATION
    if not seq.has_all_subruns():
        log.warning("At least one subrun is missing!")
        return False

    if seq.is_flawless(no_dl2=no_dl2):
        seq.understood = True
        log.info("Is flawless")
        seq.readyToClose = True
        return True

    return False


def main():
    """Check for job completion and close sequence if necessary."""
    args = autocloser_cli_parser().parse_args()

    # for the console output
    log.setLevel(logging.INFO)

    if args.log:
        fh = logging.FileHandler(args.log)
        fh.setLevel(logging.DEBUG)
        log.addHandler(fh)
        log.info(f"Logging verbose output to {args.log}")

    if args.verbose:
        log.setLevel(logging.DEBUG)
        log.debug("Verbose output.")

    if args.simulate:
        log.debug("Simulation mode")

    if args.test:
        log.debug("Test mode.")

    if args.ignore_cronlock:
        log.debug("Ignoring cron.lock")

    if args.force:
        log.debug("Force the closing")

    if args.runwise:
        log.debug("Closing run-wise")

    if args.no_dl2:
        log.debug("Assumed no DL2 production")

    if args.date:
        options.date = args.date
        hour = 12
    else:
        options.date = datetime.datetime.now()
        hour = datetime.datetime.now().hour

    options.tel_id = args.tel_id
    options.prod_id = get_prod_id()
    date = date_to_iso(options.date)

    log.info(
        f"========== Starting {Path(__file__).stem} at "
        f"{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
        f" for date {date} ==========="
    )

    if is_night_time(hour):
        sys.exit(1)

    elif is_day_closed():
        log.info(f"Date {date} already closed for {options.tel_id}")
        sys.exit(0)

    # create telescope and sequence objects
    log.info("Simulating sequencer...")

    telescope = Telescope(args.tel_id, date, args.config)

    log.info(f"Processing {args.tel_id}...")

    # Loop over the sequences
    for sequence in telescope:
        log.info(f"Processing sequence {sequence.dict_sequence['Run']}...")

        if not understand_sequence(sequence, no_dl2=args.no_dl2):
            log.warning(f"Could not interpret sequence {sequence.dict_sequence['Run']}")
            continue

        if args.runwise and sequence.readyToClose and not sequence.closed:
            sequence.close(
                date=date,
                config=args.config,
                no_dl2=args.no_dl2,
                simulate=args.simulate,
                test=args.test,
            )

    # skip these checks if closing is forced
    if not args.force and not all(seq.readyToClose for seq in telescope):
        sys.exit(f"{args.tel_id} is NOT ready to close. Exiting.")

    log.info(f"Closing {args.tel_id}...")

    if not telescope.close(
        date=date, config=args.config, no_dl2=args.no_dl2, simulate=args.simulate, test=args.test
    ):
        log.warning(f"Could not close the day for {args.tel_id}!")
        # Send email, if later than 18:00 UTC and telescope is not ready to close
        if hour > 14:
            send_warning_mail(date=date)

    log.info("Exit")


if __name__ == "__main__":
    main()



"""Provenance i/o conversion functions."""

import datetime
from pathlib import Path

import yaml
from prov.dot import prov_to_dot
from prov.model import ProvDocument

from osa.provenance.utils import get_log_config

# config
CONFIG_PATH = Path(__file__).resolve().parent / "config"
# LOGGER_FILE = CONFIG_PATH / "logger.yaml"
provconfig = yaml.safe_load(get_log_config())
PROV_PREFIX = provconfig["PREFIX"]
DEFAULT_NS = "id"  # "logprov"

__all__ = ["provlist2provdoc", "provdoc2graph", "provdoc2json", "read_prov"]


def provlist2provdoc(provlist):
    """
    Convert a list of provenance dictionaries to
    a provdoc W3C PROV compliant structure.
    """

    pdoc = ProvDocument()
    pdoc.set_default_namespace("param:")
    pdoc.add_namespace(DEFAULT_NS, DEFAULT_NS + ":")
    records = {}
    for provdict in provlist:
        if "session_id" in provdict:
            sess_id = DEFAULT_NS + ":" + str(provdict.pop("session_id"))
            if sess_id in records:
                sess = records[sess_id]
            else:
                sess = pdoc.entity(sess_id)
                records[sess_id] = sess
            sess.add_attributes(
                {
                    "prov:label": provdict.pop("name"),
                    "prov:type": "ExecutionSession",
                    "prov:generatedAtTime": provdict.pop("startTime"),
                    "software_version": provdict.pop("software_version"),
                    "observation_date": provdict.pop("observation_date"),
                    "observation_run": provdict.pop("observation_run"),
                    "config_file": provdict.pop("config_file"),
                    "config_file_hash": provdict.pop("config_file_hash"),
                    "config_file_hash_type": provdict.pop("config_file_hash_type"),
                }
            )
        # activity
        if "activity_id" in provdict:
            act_id = DEFAULT_NS + ":" + str(provdict.pop("activity_id")).replace("-", "")
            if act_id in records:
                act = records[act_id]
            else:
                act = pdoc.activity(act_id)
                records[act_id] = act
            # activity name
            if "name" in provdict:
                act.add_attributes({"prov:label": provdict.pop("name")})
            if "script" in provdict:
                act.add_attributes({"script": provdict.pop("script")})
            # activity start
            if "startTime" in provdict:
                act.set_time(startTime=datetime.datetime.fromisoformat(provdict.pop("startTime")))
            # activity end
            if "endTime" in provdict:
                act.set_time(endTime=datetime.datetime.fromisoformat(provdict.pop("endTime")))
            # in session?
            # if "in_session" in provdict:
            #     sess_id = DEFAULT_NS + ":" + str(provdict.pop("in_session"])
            #     pdoc.wasInfluencedBy(
            #         act_id, sess_id
            #     )  # , other_attributes={'prov:type': "Context"})
            # activity configuration
            if "agent_name" in provdict:
                agent_id = str(provdict.pop("agent_name"))
                if ":" not in agent_id:
                    agent_id = DEFAULT_NS + ":" + agent_id
                else:
                    new_ns = agent_id.split(":").pop(0)
                    pdoc.add_namespace(new_ns, new_ns + ":")
                if agent_id in records:
                    agent = records[agent_id]
                else:
                    agent = pdoc.agent(agent_id)
                    records[agent_id] = agent
                # act.wasAssociatedWith(agent, attributes={"prov:role": "Creator"})
            if "parameters" in provdict:
                params_record = provdict.pop("parameters")
                params = {k: str(params_record[k]) for k in params_record}
                par = pdoc.entity(act_id + "_parameters", other_attributes=params)
                par.add_attributes({"prov:type": "Parameters"})
                par.add_attributes({"prov:label": "Parameters"})
                act.used(par, attributes={"prov:type": "Setup"})
            # usage
            if "used_id" in provdict:
                ent_id = str(provdict.pop("used_id"))
                if ":" not in ent_id:
                    ent_id = DEFAULT_NS + ":" + ent_id
                else:
                    new_ns = ent_id.split(":").pop(0)
                    pdoc.add_namespace(new_ns, new_ns + ":")
                if ent_id in records:
                    ent = records[ent_id]
                else:
                    ent = pdoc.entity(ent_id)
                    records[ent_id] = ent
                rol = provdict.pop("used_role", None)
                # if rol:
                #     ent.add_attributes({'prov:label': rol})
                act.used(ent, attributes={"prov:role": rol})
            # generation
            if "generated_id" in provdict:
                ent_id = str(provdict.pop("generated_id"))
                if ":" not in ent_id:
                    ent_id = DEFAULT_NS + ":" + ent_id
                else:
                    new_ns = ent_id.split(":").pop(0)
                    pdoc.add_namespace(new_ns, new_ns + ":")
                if ent_id in records:
                    ent = records[ent_id]
                else:
                    ent = pdoc.entity(ent_id)
                    records[ent_id] = ent
                rol = provdict.pop("generated_role", None)
                # if rol:
                #     ent.add_attributes({'prov:label': rol})
                ent.wasGeneratedBy(act, attributes={"prov:role": rol})
            for k, v in provdict.items():
                if k != "session_tag":
                    act.add_attributes({k: str(v)})
        # entity
        if "entity_id" in provdict:
            ent_id = str(provdict.pop("entity_id"))
            if ":" not in ent_id:
                ent_id = DEFAULT_NS + ":" + ent_id
            else:
                new_ns = ent_id.split(":").pop(0)
                pdoc.add_namespace(new_ns, new_ns + ":")
            if ent_id in records:
                ent = records[ent_id]
            else:
                ent = pdoc.entity(ent_id)
                records[ent_id] = ent
            if "name" in provdict:
                ent.add_attributes({"prov:label": provdict.pop("name")})
            if "type" in provdict:
                ent.add_attributes({"prov:type": provdict.pop("type")})
            if "value" in provdict:
                ent.add_attributes({"prov:value": str(provdict.pop("value"))})
            if "location" in provdict:
                ent.add_attributes({"prov:location": str(provdict.pop("location"))})
            # member
            if "member_id" in provdict:
                mem_id = str(provdict.pop("member_id"))
                if ":" not in mem_id:
                    mem_id = DEFAULT_NS + ":" + mem_id
                else:
                    new_ns = mem_id.split(":").pop(0)
                    pdoc.add_namespace(new_ns, new_ns + ":")
                if mem_id in records:
                    mem = records[mem_id]
                else:
                    mem = pdoc.entity(mem_id)
                    records[mem_id] = mem
                ent.hadMember(mem)
            if "progenitor_id" in provdict:
                progen_id = str(provdict.pop("progenitor_id"))
                if ":" not in progen_id:
                    progen_id = DEFAULT_NS + ":" + progen_id
                else:
                    new_ns = progen_id.split(":").pop(0)
                    pdoc.add_namespace(new_ns, new_ns + ":")
                if progen_id in records:
                    progen = records[progen_id]
                else:
                    progen = pdoc.entity(progen_id)
                    records[progen_id] = progen
                ent.wasDerivedFrom(progen)
            for k, v in provdict.items():
                if k not in ["session_tag", "hash", "hash_type"]:
                    ent.add_attributes({k: str(v)})
        # agent
    return pdoc


def provdoc2graph(provdoc, filename, fmt):
    """Create a graph of a provenance workflow session."""
    dot = prov_to_dot(
        provdoc,
        use_labels=True,
        show_element_attributes=True,
        show_relation_attributes=False,
    )
    content = dot.create(format=fmt)
    with open(filename, "wb") as f:
        f.write(content)


def provdoc2json(provdoc, filename):
    json_stream = provdoc.serialize(indent=4)
    with open(filename, "w") as f:
        f.write(json_stream)


def read_prov(filename="prov.log", start=None, end=None):
    """Read a list of provenance dictionaries from the logfile."""
    start_dt = datetime.datetime.fromisoformat(start) if start else None
    end_dt = datetime.datetime.fromisoformat(end) if end else None
    prov_list = []
    with open(filename, "r") as f:
        for line in f.readlines():
            ll = line.split(PROV_PREFIX)
            if len(ll) >= 2:
                prov_str = ll.pop()
                prov_dt = datetime.datetime.fromisoformat(ll.pop())
                keep = True
                if prov_dt:
                    if start and prov_dt < start_dt:
                        keep = False
                    if end and prov_dt > end_dt:
                        keep = False
                if keep:
                    prov_dict = yaml.safe_load(prov_str)
                    prov_list.append(prov_dict)
    return prov_list



"""Script to reprocess a list of dates without overwhelming the job system manager."""

import logging
import subprocess as sp
import time
from pathlib import Path

import click

from osa.configs.config import DEFAULT_CFG
from osa.utils.logging import myLogger
from osa.utils.utils import wait_for_daytime

log = myLogger(logging.getLogger(__name__))


def number_of_pending_jobs():
    """Return the number of jobs in the slurm queue."""
    cmd = ["squeue", "-u", "lstanalyzer", "-h", "-t", "pending", "-r"]
    output = sp.check_output(cmd)
    return output.count(b"\n")


def run_script(
    script: str, date, config: Path, no_dl2: bool, no_calib: bool, simulate: bool, force: bool
):
    """Run the sequencer for a given date."""
    osa_config = Path(config).resolve()

    cmd = [script, "--config", str(osa_config), "--date", date]

    if no_dl2:
        cmd.append("--no-dl2")

    if no_calib:
        cmd.append("--no-calib")

    if simulate:
        cmd.append("--simulate")

    if force:
        cmd.append("--force")

    # Append the telescope to the command in the last place
    cmd.append("LST1")

    log.info(f"\nRunning {' '.join(cmd)}")
    sp.run(cmd)


def check_job_status_and_wait(max_jobs=2500):
    """Check the status of the jobs in the queue and wait for them to finish."""
    while number_of_pending_jobs() > max_jobs:
        log.info("Waiting 2 hours for slurm queue to decrease...")
        time.sleep(7200)


def get_list_of_dates(dates_file):
    """Read the files with the dates to be processed and build a list of dates."""
    with open(dates_file, "r") as file:
        list_of_dates = [line.strip() for line in file]
    return list_of_dates


@click.command()
@click.option("--no-dl2", is_flag=True, help="Do not run the DL2 step.")
@click.option("--no-calib", is_flag=True, help="Do not run the calibration step.")
@click.option("-s", "--simulate", is_flag=True, help="Activate simulation mode.")
@click.option("-f", "--force", is_flag=True, help="Force the autocloser to close the day.")
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True),
    default=DEFAULT_CFG,
    help="Path to the OSA config file.",
)
@click.argument(
    "script", type=click.Choice(["sequencer", "closer", "copy_datacheck", "autocloser"])
)
@click.argument("dates-file", type=click.Path(exists=True))
def main(
    script: str = None,
    dates_file: Path = None,
    config: Path = DEFAULT_CFG,
    no_dl2: bool = False,
    no_calib: bool = False,
    simulate: bool = False,
    force: bool = False,
):
    """
    Loop over the dates listed in the input file and launch the script for each of them.
    The input file should list the dates in the format YYYY-MM-DD one date per line.
    """
    logging.basicConfig(level=logging.INFO)

    list_of_dates = get_list_of_dates(dates_file)

    # Check slurm queue status
    check_job_status_and_wait()

    for date in list_of_dates:
        # Avoid running jobs while it is still night time
        wait_for_daytime()

        run_script(script, date, config, no_dl2, no_calib, simulate, force)
        log.info("Waiting 1 minute to launch the process for the next date...\n")
        time.sleep(60)

        # Check slurm queue status and sleep for a while to avoid overwhelming the queue
        check_job_status_and_wait()

    log.info("Done! No more dates to process.")


if __name__ == "__main__":
    main()



"""Simulate executions of data processing pipeline and produce provenance.

If it is not executed by tests, please run  pytest --basetemp=test_osa first.
It needs to have test_osa folder filled with test datasets.

python osa/scripts/simulate_processing.py"""

import logging
import multiprocessing as mp
import subprocess
from datetime import datetime
from pathlib import Path

import yaml

from osa.configs import options
from osa.configs.config import cfg
from osa.job import calibration_sequence_job_template, data_sequence_job_template
from osa.nightsummary.extract import build_sequences
from osa.provenance.utils import get_log_config
from osa.utils.cliopts import simprocparsing
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_dir

__all__ = [
    "parse_template",
    "do_setup",
    "simulate_processing",
    "simulate_subrun_processing",
    "tear_down",
]

CONFIG_FLAGS = {
    "Go": True,
    "TearDL1": False,
    "TearDL2": False,
    "TearSubDL1": False,
    "TearSubDL2": False,
    "TearAnalysis": False,
    "TearSubAnalysis": False,
}
provconfig = yaml.safe_load(get_log_config())
LOG_FILENAME = provconfig["handlers"]["provHandler"]["filename"]

log = myLogger(logging.getLogger())


def do_setup():
    """Set-up folder structure and check flags."""
    path_analysis = Path(cfg.get("LST1", "ANALYSIS_DIR")) / options.directory
    path_dl1 = Path(cfg.get("LST1", "DL1_DIR")) / options.directory
    path_dl2 = Path(cfg.get("LST1", "DL2_DIR")) / options.directory
    path_sub_analysis = path_analysis / options.prod_id
    path_dl1sub = path_dl1 / options.prod_id
    path_dl2_sub = path_dl2 / options.prod_id

    if Path(LOG_FILENAME).exists() and not options.append:
        CONFIG_FLAGS["Go"] = False
        log.info(f"File {LOG_FILENAME} already exists.")
        log.info(f"You must rename/remove {LOG_FILENAME} to produce a clean provenance.")
        log.info("You can also set --append flag to append captured provenance.")
        return

    CONFIG_FLAGS["TearSubAnalysis"] = (
        False if path_sub_analysis.exists() or options.provenance else path_sub_analysis
    )
    CONFIG_FLAGS["TearAnalysis"] = (
        False if path_analysis.exists() or options.provenance else path_analysis
    )
    CONFIG_FLAGS["TearSubDL1"] = (
        False if path_dl1sub.exists() or options.provenance else path_dl1sub
    )
    CONFIG_FLAGS["TearSubDL2"] = (
        False if path_dl2_sub.exists() or options.provenance else path_dl2_sub
    )
    CONFIG_FLAGS["TearDL1"] = False if path_dl1.exists() or options.provenance else path_dl1
    CONFIG_FLAGS["TearDL2"] = False if path_dl2.exists() or options.provenance else path_dl2

    if options.provenance and not options.force:
        if path_sub_analysis.exists():
            CONFIG_FLAGS["Go"] = False
            log.info(f"Folder {path_sub_analysis} already exist.")
        if path_dl1sub.exists():
            CONFIG_FLAGS["Go"] = False
            log.info(f"Folder {path_dl1sub} already exist.")
        if path_dl2_sub.exists():
            CONFIG_FLAGS["Go"] = False
            log.info(f"Folder {path_dl2_sub} already exist.")
        if not CONFIG_FLAGS["Go"]:
            log.info("You must enforce provenance files overwrite with --force flag.")
            return

    path_sub_analysis.mkdir(parents=True, exist_ok=True)
    path_dl1sub.mkdir(parents=True, exist_ok=True)
    path_dl2_sub.mkdir(parents=True, exist_ok=True)


def tear_down():
    """Tear down created temporal folders."""
    if isinstance(CONFIG_FLAGS["TearSubAnalysis"], Path):
        CONFIG_FLAGS["TearSubAnalysis"].rmdir()
    if isinstance(CONFIG_FLAGS["TearAnalysis"], Path):
        CONFIG_FLAGS["TearAnalysis"].rmdir()
    if isinstance(CONFIG_FLAGS["TearSubDL1"], Path):
        CONFIG_FLAGS["TearSubDL1"].rmdir()
    if isinstance(CONFIG_FLAGS["TearSubDL2"], Path):
        CONFIG_FLAGS["TearSubDL2"].rmdir()
    if isinstance(CONFIG_FLAGS["TearDL1"], Path):
        CONFIG_FLAGS["TearDL1"].rmdir()
    if isinstance(CONFIG_FLAGS["TearDL2"], Path):
        CONFIG_FLAGS["TearDL2"].rmdir()


def parse_template(template, idx):
    """Parse batch templates."""
    args = []
    keep = False
    for line in template.splitlines():
        if keep:
            line = line.replace("f'", "")
            line = line.replace("'", "")
            line = line.replace(",", "")
            line = line.replace(r"{subruns:04d}", str(idx).zfill(4))
            args.append(line.strip())
        if "subprocess.run" in line:
            keep = True
    # remove last three elements
    return args[:-3]


def simulate_calibration(args):
    """Simulate calibration."""
    log.info("Simulating calibration call")
    subprocess.run(args)


def simulate_subrun_processing(args):
    """Simulate subrun processing."""
    run_str, subrun_idx = args[-2].split(".")
    log.info(f"Simulating process call for run {run_str} subrun {subrun_idx}")
    subprocess.run(args)


def simulate_processing():
    """Simulate daily processing and capture provenance."""
    options.simulate = True
    options.test = True

    sequence_list = build_sequences(options.date)

    # simulate data calibration and reduction
    for sequence in sequence_list:
        processed = False
        if sequence.type == "PEDCALIB":
            args_cal = parse_template(calibration_sequence_job_template(sequence), 0)
            simulate_calibration(args_cal)
        elif sequence.type == "DATA":
            with mp.Pool() as poolproc:
                args_proc = [
                    parse_template(data_sequence_job_template(sequence), subrun_idx)
                    for subrun_idx in range(sequence.subruns)
                ]
                processed = poolproc.map(simulate_subrun_processing, args_proc)
        # produce prov if overwrite prov arg
        if processed and options.provenance:
            command = "provprocess"
            drs4_pedestal_run_id = f"{sequence.drs4_run:05d}"
            pedcal_run_id = f"{sequence.pedcal_run:05d}"

            args_pp = [
                command,
                "-c",
                f"{options.configfile}",
                drs4_pedestal_run_id,
                pedcal_run_id,
                sequence.run_str,
                options.directory,
                options.prod_id,
            ]
            log.info(f"Processing provenance for run {sequence.run_str}")
            subprocess.run(args_pp, check=True)


def main():
    """Dry-run of the entire processing chain to produce provenance."""
    log.setLevel(logging.INFO)

    simprocparsing()

    # date and tel_id hardcoded for the moment
    options.date = datetime.fromisoformat("2020-01-17")
    options.tel_id = "LST1"
    options.directory = date_to_dir(options.date)

    log.info("Running simulate processing")

    do_setup()
    if CONFIG_FLAGS["Go"]:
        simulate_processing()
    tear_down()


if __name__ == "__main__":
    main()



import logging
import subprocess as sp
from datetime import datetime
from pathlib import Path
from textwrap import dedent

import click
import pandas as pd
from astropy import units as u
from astropy.table import Table, join, unique, vstack
from astropy.time import Time
from lstchain.io.io import dl1_params_lstcam_key
from lstchain.reco.utils import add_delta_t_key, get_effective_time

from osa.paths import get_major_version
from osa.utils.utils import get_lstchain_version

pd.set_option("display.float_format", "{:.1f}".format)

logging.basicConfig(level=logging.INFO, format="%(asctime)s:%(levelname)s:%(message)s")
log = logging.getLogger(__name__)


BASE_DL1 = Path("/fefs/aswg/data/real/DL1")
BASE_MONITORING = Path("/fefs/aswg/data/real/monitoring")
CATALOG_DIR = Path("/fefs/aswg/data/real/OSA/Catalog")


def add_table_to_html(html_table):
    return dedent(
        f"""\
    <html>
      <head>
        <link href="osa.css" rel="stylesheet" type="text/css">
      </head>
    <body>
    {html_table}
    </body>
    </html>
    """
    )


def add_query_table_to_html(html_table):
    return dedent(
        f"""\
    <html>
     <head>
        <meta http-equiv="Content-type" content="text/html; charset=utf-8">
    <style>
    body {{font-family: sans-serif;}}
    table.dataTable {{width: auto !important; margin: 0 !important;}}
    .dataTables_filter, .dataTables_paginate {{float: left !important; margin-left:1em}}
    </style>
        <link rel="stylesheet" type="text/css" href="https://cdn.datatables.net/1.11.5/css/jquery.dataTables.min.css">
        <script type="text/javascript" language="javascript" src="https://code.jquery.com/jquery-3.5.1.js"></script>
        <script type="text/javascript" language="javascript" src="https://cdn.datatables.net/1.11.5/js/jquery.dataTables.min.js"></script>

    <script type="text/javascript" class="init">

    $(document).ready(function() {{
        $('#example').DataTable();
    }} );

    </script>

    </head>

     <body>
      <script>
    var astropy_sort_num = function(a, b) {{
        var a_num = parseFloat(a);
        var b_num = parseFloat(b);

        if (isNaN(a_num) && isNaN(b_num))
            return ((a < b) ? -1 : ((a > b) ? 1 : 0));
        else if (!isNaN(a_num) && !isNaN(b_num))
            return ((a_num < b_num) ? -1 : ((a_num > b_num) ? 1 : 0));
        else
            return isNaN(a_num) ? -1 : 1;
    }}

    jQuery.extend( jQuery.fn.dataTableExt.oSort, {{
        "optionalnum-asc": astropy_sort_num,
        "optionalnum-desc": function (a,b) {{ return -astropy_sort_num(a, b); }}
    }});

    $(document).ready(function() {{
        $('#table139855676982704').dataTable({{
            order: [],
            pageLength: 50,
            lengthMenu: [[10, 25, 50, 100, -1], [10, 25, 50, 100, 'All']],
            pagingType: "full_numbers",
            columnDefs: [{{targets: [0, 1, 2, 4], type: "optionalnum"}}]
        }});
    }} );
    </script>
        {html_table}
        </body>
        </html>
        """
    )


def add_start_and_elapsed(table: Table, datedir: str, version: str) -> None:
    """Add columns with the timestamp of first events and elapsed time of the runs.

    This information is taken from the merged DL1 files. Two new columns are added
    to the input table.

    Parameters
    ----------
    table : astropy.table.Table
        Astropy Table to which the two new columns are to be added.
    datedir : str
        Date directory in YYYYMMDD format.
    version : str
        Production version of the processing in the format 'vW.X.Y.Z'.
    """
    if "run_id" not in table.columns:
        raise KeyError("Run ID not present in given table. Please check its content.")

    start_times = []
    elapsed_times = []

    for run in table["run_id"]:
        major_version = get_major_version(version)
        file = BASE_DL1 / datedir / major_version / f"tailcut84/dl1_LST-1.Run{run:05d}.h5"
        df = pd.read_hdf(file, key=dl1_params_lstcam_key)

        # Timestamp of the first event
        first_time = Time(df["dragon_time"][0], format="unix", scale="utc")
        start_times.append(first_time.utc.iso)

        # Elapsed time of the run
        df_delta = add_delta_t_key(df)
        _, elapsed_t = get_effective_time(df_delta)
        elapsed_times.append(elapsed_t.to(u.min))

    # Modify the input table by adding two new columns
    table.add_column(elapsed_times, name="Elapsed [min]")
    table.add_column(start_times, name="Run start [UTC]")


def copy_to_webserver(html_file, csv_file):
    sp.run(["scp", str(html_file), "datacheck:/home/www/html/datacheck/lstosa/."], check=True)
    sp.run(["scp", str(csv_file), "datacheck:/home/www/html/datacheck/lstosa/."], check=True)


@click.command()
@click.argument("date", type=click.DateTime(formats=["%Y-%m-%d"]))
@click.option("-v", "--version", type=str, default=get_lstchain_version())
def main(date: datetime = None, version: str = get_lstchain_version()):
    """Update source catalog with new run entries from a given date in format YYYY-MM-DD.

    Notes
    -----
    It needs to be run as lstanalyzer user.
    """
    catalog_path = CATALOG_DIR / "LST_source_catalog.ecsv"
    catalog_table = Table.read(catalog_path)

    # Open table for given date and append its content to the table with entire catalog
    datedir = date.strftime("%Y%m%d")
    today_catalog = Table.read(BASE_MONITORING / f"RunCatalog/RunCatalog_{datedir}.ecsv")
    today_runsummary = Table.read(BASE_MONITORING / f"RunSummary/RunSummary_{datedir}.ecsv")
    # Keep only astronomical data runs
    today_runsummary = today_runsummary[today_runsummary["run_type"] == "DATA"]
    todays_info = join(today_runsummary, today_catalog)
    todays_info.add_column(date.strftime("%Y-%m-%d"), name="date_dir")
    todays_info.keep_columns(["run_id", "source_name", "date_dir"])

    # Add start of run in iso format and elapsed time for each run
    log.info("Getting run start and elapsed time")
    add_start_and_elapsed(todays_info, datedir, version)

    # Change column names
    todays_info.rename_column("run_id", "Run ID")
    todays_info.rename_column("source_name", "Source name")
    todays_info.rename_column("date_dir", "Date directory")

    # Add new rows from given date to the whole catalog table
    log.info("Adding new rows to table")
    new_table = vstack([catalog_table, todays_info])
    table_unique = unique(new_table, keys="Run ID", keep="last")

    # To pandas
    log.info("Converting to pandas and HTML")
    df = table_unique.to_pandas()
    df = df.sort_values(by="Run ID", ascending=False)

    # To HTML
    html_table = df.to_html(index=False, justify="left")
    html_table = html_table.replace(
        '<table border="1" class="dataframe">',
        '<table class="display compact" id="table139855676982704">',
    )
    html_content = add_query_table_to_html(html_table)

    # Save the HTML and ECSV files and copy them to the LST-1 webserver
    html_file = CATALOG_DIR / "LST_source_catalog.html"
    html_file.write_text(html_content)
    table_unique.write(catalog_path, delimiter=",", overwrite=True)

    copy_to_webserver(html_file, catalog_path)


if __name__ == "__main__":
    main()



"""
Manage the stages of the analysis workflow.

Build lstchain commands, run them, clean up their output of a given step and
retry in case of failure, and keep track of the history of the stages.
"""

import logging
import subprocess as sp
from pathlib import Path
from typing import List, Union

from tenacity import retry, stop_after_attempt
import lstchain

from osa.configs import options
from osa.configs.config import cfg
from osa.report import history
from osa.utils.logging import myLogger
from osa.utils.utils import stringify, date_to_dir
from osa.paths import get_run_date

log = myLogger(logging.getLogger(__name__))


class AnalysisStage:
    """Run a given analysis stage keeping track of checkpoints in a history file.

    Parameters
    ----------
    run: str
        Run number
    command_args: List[str]
        Complete command line arguments to be executed in the shell as a list
    config_file: str, optional
        Path to the config file used for the stage
    """

    def __init__(
        self,
        run: str,
        command_args: List[str],
        config_file: Union[str, None] = None,
    ):
        self.run = run
        self.command_args = command_args
        self.config_file = config_file
        self.command = self.command_args[0]
        self.rc = None
        self.history_file = (
            Path(options.directory) / f"sequence_{options.tel_id}_{self.run}.history"
        )

    @retry(stop=stop_after_attempt(int(cfg.get("lstchain", "max_tries"))))
    def execute(self):
        """Run the program and retry if it fails."""
        log.info(f"Executing {stringify(self.command_args)}")
        output = sp.run(self.command_args, stdout=sp.PIPE, stderr=sp.STDOUT, encoding="utf-8")
        self.rc = output.returncode
        self._write_checkpoint()

        # If fails, remove products from the directory for subsequent trials
        if self.rc != 0:
            self._clean_up()
            raise ValueError(f"{self.command} failed with output: \n {output.stdout}")

    def show_command(self):
        """Show the command to be executed."""
        return stringify(self.command_args)

    def _clean_up(self):
        """
        Clean up the output files created at a given analysis stage that exited with a
        non-zero return code, provided the output file exists. In this way, they can be
        reproduced in subsequent trials.
        """

        if self.command == "lstchain_data_r0_to_dl1":
            self._remove_dl1a_output()
        elif self.command == "lstchain_dl1ab":
            self._remove_dl1b_output('dl1_LST-1.Run')
        elif self.command == "lstchain_check_dl1":
            self._remove_dl1b_output('datacheck_dl1_LST-1.Run')
        elif self.command == "onsite_create_calibration_file":
            self._remove_calibration()
        elif self.command == "onsite_create_drs4_pedestal_file":
            self._remove_drs4_baseline()

    def _remove_drs4_baseline(self):
        drs4_pedestal_basedir = Path(cfg.get("LST1", "PEDESTAL_DIR"))
        date = date_to_dir(get_run_date(self.run))
        drs4_pedestal_dir = drs4_pedestal_basedir / date / lstchain.__version__
        file = drs4_pedestal_dir / "drs4_pedestal.Run{self.run}.0000.h5"
        file.unlink(missing_ok=True)
        # Also remove the link to "pro" directory
        drs4_pedestal_dir_pro = drs4_pedestal_basedir / date / "pro"
        drs4_pedestal_dir_pro.unlink(missing_ok=True)

    def _remove_calibration(self):
        calib_basedir = Path(cfg.get("LST1", "CALIB_DIR"))
        date = date_to_dir(get_run_date(self.run))
        calib_dir = file = calib_basedir / date / lstchain.__version__
        file = calib_dir / f"calibration_filters_{options.filters}.Run{self.run}.0000.h5"
        file.unlink(missing_ok=True)
        # Also remove the link to "pro" directory
        calib_dir_pro = file = calib_basedir / date / "pro"
        calib_dir_pro.unlink(missing_ok=True)

    def _remove_dl1a_output(self):
        dl1_output_file = options.directory / f"dl1_LST-1.Run{self.run}.h5"
        muon_output_file = options.directory / f"muons_LST-1.Run{self.run}.fits"
        interleaved_output_file = options.directory / f"interleaved_LST-1.Run{self.run}.h5"
        dl1_output_file.unlink(missing_ok=True)
        muon_output_file.unlink(missing_ok=True)
        interleaved_output_file.unlink(missing_ok=True)

    def _remove_dl1b_output(self, file_prefix):
        dl1ab_subdirectory = options.directory / options.dl1_prod_id
        output_file = dl1ab_subdirectory / f"{file_prefix}{self.run}.h5"
        output_file.unlink(missing_ok=True)

    def _write_checkpoint(self):
        """Write the checkpoint in the history file."""
        history(
            run=self.run,
            prod_id=options.prod_id,
            stage=self.command,
            return_code=self.rc,
            history_file=self.history_file,
            config_file=self.config_file,
        )


class DRS4PedestalStage(AnalysisStage):
    """
    Class inheriting from AnalysisStage for the first part of the calibration procedure,
    i.e. the pedestal subtraction.

    Notes
    -----
    The history file is the same for both calibration stages and carries the name of
    the pedcal run.
    """

    def __init__(
        self,
        run: str,
        run_pedcal: str,
        command_args: List[str],
        config_file: Union[str, None] = None,
    ):
        super().__init__(run, command_args, config_file)
        self.run_pedcal = run_pedcal
        self.history_file = (
            Path(options.directory) / f"sequence_{options.tel_id}_{self.run_pedcal}.history"
        )

    def _write_checkpoint(self):
        """Write the checkpoint in the history file."""
        history(
            run=self.run,
            prod_id=options.prod_id,
            stage=self.command,
            return_code=self.rc,
            history_file=self.history_file,
        )


class ChargeCalibrationStage(AnalysisStage):
    """Class inheriting from AnalysisStage for the second part of the calibration procedure."""

    def __init__(
        self,
        run: str,
        command_args: List[str],
        config_file: Union[str, None] = None,
    ):
        super().__init__(run, command_args, config_file)
        self.history_file = (
            Path(options.directory) / f"sequence_{options.tel_id}_{self.run}.history"
        )

    def _write_checkpoint(self):
        """Write the checkpoint in the history file."""
        history(
            run=self.run,
            prod_id=options.prod_id,
            stage=self.command,
            return_code=self.rc,
            history_file=self.history_file,
        )



"""Tweaks for logging module. Custom formatter for the different logging levels."""

import logging

DEFAULT_LOGGING_FORMAT = (
    "%(asctime)s %(levelname)s [%(name)s] (%(module)s.%(funcName)s): %(message)s"
)

__all__ = ["myLogger"]


class MyFormatter(logging.Formatter):
    """Customize formatter of info logging level."""

    info_fmt = "%(message)s"

    def __init__(self):
        super().__init__(fmt=DEFAULT_LOGGING_FORMAT, datefmt=None, style="%")

    def format(self, record):
        """
        Parameters
        ----------
        record: logging record

        Returns
        -------
        result: logging formatter

        """
        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._style._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.INFO:
            self._style._fmt = MyFormatter.info_fmt

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = format_orig

        return result


def myLogger(osalogger):
    """Creates a logger with a customized formatted handler."""
    fmt = MyFormatter()
    handler = logging.StreamHandler()
    handler.setFormatter(fmt)
    osalogger.addHandler(handler)
    osalogger.propagate = False
    return osalogger



from datetime import datetime
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.utils import date_to_dir

options.date = datetime.fromisoformat("2020-01-17")
options.prod_id = "v0.1.1"


def test_get_calibration_file(r0_data, merged_run_summary):
    from osa.paths import get_calibration_filename

    for file in r0_data:
        assert file.exists()
    file = get_calibration_filename(1809, options.prod_id)
    file.exists()


def test_get_drs4_pedestal_file(r0_data, merged_run_summary):
    from osa.paths import get_drs4_pedestal_filename

    for file in r0_data:
        assert file.exists()
    file = get_drs4_pedestal_filename(1804, options.prod_id)
    file.exists()


def test_pedestal_ids_file_exists(pedestal_ids_file):
    from osa.paths import pedestal_ids_file_exists

    pedestal_ids_file.exists()
    assert pedestal_ids_file_exists(1808) is True


def test_get_datacheck_file(datacheck_dl1_files):
    from osa.paths import get_datacheck_files

    for file in datacheck_dl1_files:
        assert file.exists()
    dl1_path = Path("test_osa/test_files0/DL1/20200117/v0.1.0/tailcut84")
    files = get_datacheck_files(pattern="datacheck*.pdf", directory=dl1_path)
    expected_files = [
        dl1_path / "datacheck_dl1_LST-1.Run01808.pdf",
        dl1_path / "datacheck_dl1_LST-1.Run01807.pdf",
    ]
    assert set(files) == set(expected_files)


def test_destination_dir():
    from osa.paths import destination_dir

    datedir = date_to_dir(options.date)
    options.dl1_prod_id = cfg.get("LST1", "DL1_PROD_ID")
    options.dl2_prod_id = cfg.get("LST1", "DL2_PROD_ID")
    options.prod_id = cfg.get("LST1", "PROD_ID")
    base_directory = cfg.get("LST1", "BASE")
    base_path = Path(base_directory)

    data_types = {
        "INTERLEAVED": "DL1",
        "DL1AB": "DL1",
        "DATACHECK": "DL1",
        "MUON": "DL1",
        "DL2": "DL2",
    }

    for concept, dst_dir in data_types.items():
        directory = destination_dir(concept, create_dir=False)
        if concept == "DL1AB":
            expected_directory = (
                base_path / dst_dir / datedir / options.prod_id / options.dl1_prod_id
            )
        elif concept == "DATACHECK":
            expected_directory = (
                base_path / dst_dir / datedir / options.prod_id / options.dl1_prod_id / "datacheck"
            )
        elif concept == "MUON":
            expected_directory = (
                base_path / dst_dir / datedir / options.prod_id / "muons"
            )
        elif concept == "INTERLEAVED":
            expected_directory = (
                base_path / dst_dir / datedir / options.prod_id / "interleaved"
            )
        elif concept == "DL2":
            expected_directory = (
                base_path / dst_dir / datedir / options.prod_id / options.dl2_prod_id
            )

        assert directory == expected_directory


def test_get_run_date(merged_run_summary):
    from osa.paths import get_run_date

    assert merged_run_summary.exists()

    assert get_run_date(1808) == datetime(2020,1,17)

    assert get_run_date(1200) == datetime(2020,1,17)



import subprocess as sp


def test_significance(
    dl2_merged,
    run_summary_file,
    r0_data,
    run_catalog,
    drs4_time_calibration_files,
    systematic_correction_files,
    pedestal_ids_file,
    merged_run_summary,
):
    output = sp.run(
        ["theta2_significance", "-d", "2020-01-17", "-s", "LST1"],
        text=True,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
    )
    assert output.returncode == 0
    assert "Source: MadeUpSource, runs: [1808]" in output.stderr.splitlines()[-1]



"""Handle the writing, reading and appending strings to files."""

import filecmp
import logging
import pathlib
from os import remove, rename

from osa.configs import options
from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))

__all__ = [
    "write_to_file",
    "append_to_file",
]


def write_to_file(file: pathlib.Path, content: str):
    """Check if the file already exists and write the content in it."""
    file_temp = f"{file}.tmp"
    try:
        with open(file_temp, "w") as file_handle:
            file_handle.write(f"{content}")
    except (IOError, OSError) as e:
        log.exception(f"{e.strerror} {e.filename}")

    if file.exists() and file.is_file():
        if filecmp.cmp(file, file_temp):
            remove(file_temp)
            return False

        if options.simulate:
            remove(file_temp)
            log.debug(f"SIMULATE File {file_temp} would replace {file}." f"Deleting {file_temp}")
        else:
            try:
                rename(file_temp, file)
            except (IOError, OSError) as e:
                log.exception(f"{e.strerror} {e.filename}")
    elif options.simulate:
        log.debug(f"SIMULATE File {file_temp} would be written as {file}. Deleting {file_temp}")
    else:
        rename(file_temp, file)
    return True


def append_to_file(file: pathlib.Path, content: str) -> None:
    """
    Check if the file already exists and write the content in it.

    Parameters
    ----------
    file: pathlib.Path
        The file to write in.
    content: str
        The content to write in the file.
    """
    if file.exists() and file.is_file():
        if options.simulate:
            log.debug(f"SIMULATE File {file} would be appended")
        else:
            with open(file, "a") as file_handle:
                try:
                    file_handle.write(content)
                except IOError as e:
                    log.exception(f"{e.strerror} {e.filename}")
    else:
        write_to_file(file, content)



import logging
from datetime import datetime
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_dir

log = myLogger(logging.getLogger(__name__))

__all__ = ["get_check_raw_dir", "get_raw_dir", "is_raw_data_available"]


def get_check_raw_dir(date: datetime) -> Path:
    """Get the raw directory and check if it contains raw files."""
    raw_dir = get_raw_dir(date)
    log.debug(f"Raw directory: {raw_dir}")

    if not raw_dir.exists():
        raise IOError(f"Raw directory {raw_dir} does not exist")

    # check that it contains raw files
    files = raw_dir.glob("*.fits.fz")
    if not files:
        raise IOError(f"Empty raw directory {raw_dir}")

    return raw_dir


def get_raw_dir(date: datetime) -> Path:
    """Get the raw R0 directory."""
    night_dir = date_to_dir(date)
    r0_dir = Path(cfg.get(options.tel_id, "R0_DIR")) / night_dir
    return r0_dir if options.tel_id == "LST1" else None


def is_raw_data_available(date: datetime) -> bool:
    """Get the raw directory and check its existence."""
    answer = False
    if options.tel_id == "LST1":
        raw_dir = get_check_raw_dir(date)
        if raw_dir.exists():
            answer = True
    else:
        answer = True
    return answer



"""
Interact with the OSA database.

Deal with the OSA database (observations, analysis, transfers).
"""

import logging
import sqlite3
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))


__all__ = ["start_processing", "end_processing", "open_database"]


@contextmanager
def open_database(filename: str):
    """Open a database connection using as a context manager."""
    if not Path(filename).exists():
        log.warning(f"Database file '{filename}' not found.")
        yield None
        return

    connection = sqlite3.connect(filename)

    try:
        yield connection.cursor()
    finally:
        connection.commit()
        connection.close()


def start_processing(date: str) -> None:
    """Indicate the start of the processing for a given date in the OSA database."""
    if options.test or options.simulate:
        return

    t_start = datetime.now()
    finished = False

    database = cfg.get("database", "path")

    with open_database(database) as cursor:
        if cursor is not None:
            cursor.execute("SELECT * FROM processing WHERE date = ?", (date,))

            if cursor.fetchone() is None:
                cursor.execute(
                    "INSERT INTO processing (telescope, date, start, prod_id, is_finished) "
                    "VALUES (?, ?, ?, ?, ?)",
                    (
                        options.tel_id,
                        date,
                        t_start,
                        options.prod_id,
                        finished,
                    ),
                )


def end_processing(date: str) -> None:
    """Indicate the end of the processing for a given date in the OSA database."""
    if options.test or options.simulate:
        return

    database = cfg.get("database", "path")
    finished = True
    t_end = datetime.now()

    # Once the night processing finishes, update the last two columns ('end' and
    # 'is_finished') of the corresponding row of a given date in the processing table.
    with open_database(database) as cursor:
        if cursor is not None:
            cursor.execute(
                "UPDATE processing SET is_finished = ?, end = ? WHERE date = ?",
                (finished, t_end, date),
            )



"""Utility functions for dealing with the copy of datacheck files to webserver."""

import logging
import subprocess as sp
from pathlib import Path
from typing import List

from osa.configs.config import cfg
from osa.paths import DATACHECK_WEB_BASEDIR
from osa.utils.logging import myLogger

__all__ = ["directory_in_webserver", "copy_to_webserver"]

log = myLogger(logging.getLogger(__name__))


def directory_in_webserver(host: str, datacheck_type: str, date: str, prod_id: str) -> Path:
    """
    Create directories in the datacheck web server.

    Parameters
    ----------
    host : str
        Hostname of the server to which the datacheck products will be copied
    datacheck_type : str
        Type of datacheck product (PEDESTAL, CALIB, DL1, LONGTERM)
    date : str
        Date in the format YYYYMMDD
    prod_id : str
        Production ID

    Returns
    -------
    Path
        Path to the directory in the web server
    """
    DATACHECK_WEB_DIRS = {
        "PEDESTAL": f"drs4/{prod_id}/{date}",
        "CALIB": f"enf_calibration/{prod_id}/{date}",
        "DL1AB": f"dl1/{prod_id}/{date}/pdf",
        "LONGTERM": f"dl1/{prod_id}/{date}",
        "HIGH_LEVEL": f"high_level/{prod_id}/{date}",
    }

    destination_dir = DATACHECK_WEB_BASEDIR / DATACHECK_WEB_DIRS[datacheck_type]

    remote_mkdir = ["ssh", host, "mkdir", "-p", destination_dir]
    sp.run(remote_mkdir, check=True)

    return destination_dir


def copy_to_webserver(files: List[Path], datacheck_type: str, date: str, prod_id: str) -> None:
    """
    Copy files to the webserver in host.

    Parameters
    ----------
    files : List[Path]
        List of files to be copied
    datacheck_type : str
        Type of datacheck product
    date : str
        Date in the format YYYYMMDD
    prod_id : str
        Production ID
    """
    host = cfg.get("WEBSERVER", "HOST")
    destination_dir = directory_in_webserver(host, datacheck_type, date, prod_id)

    for file in files:
        log.info(f"Copying {file}")
        copy_file = ["scp", file, f"{host}:{destination_dir}/."]
        sp.run(copy_file, check=True, stdout=sp.DEVNULL, stderr=sp.DEVNULL)



"""Handle the creation and reading of the run summary files."""

import logging
import subprocess
from pathlib import Path

from astropy.table import Table

from osa.configs.config import cfg
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_dir, stringify

__all__ = ["produce_run_summary_file", "get_run_summary_file", "run_summary_table"]


log = myLogger(logging.getLogger(__name__))


def produce_run_summary_file(date) -> None:
    """
    Produce the run summary using the lstchain script.

    Parameters
    ----------
    date : datetime.datetime
    """
    nightdir = date_to_dir(date)
    r0_dir = Path(cfg.get("LST1", "R0_DIR"))
    run_summary_dir = Path(cfg.get("LST1", "RUN_SUMMARY_DIR"))

    command = [
        "lstchain_create_run_summary",
        f"--date={nightdir}",
        f"--r0-path={r0_dir}",
        f"--output-dir={run_summary_dir}",
    ]

    try:
        subprocess.run(command, check=True)

    except OSError as error:
        log.exception(f"Command {stringify(command)}, error: {error}")
    except subprocess.CalledProcessError as error:
        log.exception(f"Subprocess error: {error}")


def run_summary_table(date) -> Table:
    """
    Reads the run summary ECSV file containing as a Table with the following content:
     - run_id, datatype: int64
     - n_subruns, datatype: int64
     - run_type, datatype: string
     - ucts_timestamp, datatype: int64
     - run_start, datatype: int64
     - dragon_reference_time, datatype: int64
     - dragon_reference_module_id, datatype: int16
     - dragon_reference_module_index, datatype: int16
     - dragon_reference_counter, datatype: uint64
     - dragon_reference_source, datatype: string

    Parameters
    ----------
    date : datetime.datetime

    Returns
    -------
    table : astropy.Table
        Table with the content of the run summary ECSV file.
    """
    night_summary_file = get_run_summary_file(date)
    log.debug(f"Looking for run summary file {night_summary_file}")

    if not night_summary_file.exists():
        log.info(f"Run summary file {night_summary_file} not found. Producing it.")
        produce_run_summary_file(date)

    table = Table.read(night_summary_file)
    table.add_index(["run_id"])
    return table


def get_run_summary_file(date) -> Path:
    """
    Builds the file name of the run summary ECSV file.

    Parameters
    ----------
    date : datetime.datetime

    Returns
    -------
    run_summary_file : pathlib.Path
        File name of the run summary ECSV file
    """
    nightdir = date_to_dir(date)
    return Path(cfg.get("LST1", "RUN_SUMMARY_DIR")) / f"RunSummary_{nightdir}.ecsv"



"""Handle the paths of the analysis products."""

import logging
import re
from datetime import datetime
from pathlib import Path
from typing import List

import lstchain
from astropy.table import Table
from lstchain.onsite import (find_systematics_correction_file,
                             find_time_calibration_file,
                             find_filter_wheels)

from osa.configs import options
from osa.configs.config import DEFAULT_CFG, cfg
from osa.configs.datamodel import Sequence
from osa.utils import utils
from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))

__all__ = [
    "get_calibration_filename",
    "get_drs4_pedestal_filename",
    "pedestal_ids_file_exists",
    "get_run_date",
    "drs4_pedestal_exists",
    "calibration_file_exists",
    "sequence_calibration_files",
    "destination_dir",
    "datacheck_directory",
    "get_datacheck_files",
    "get_drive_file",
    "get_summary_file",
    "get_pedestal_ids_file",
    "DATACHECK_WEB_BASEDIR",
    "DEFAULT_CFG",
    "create_source_directories",
    "analysis_path",
]


DATACHECK_WEB_BASEDIR = Path(cfg.get("WEBSERVER", "DATACHECK"))
CALIB_BASEDIR = Path(cfg.get("LST1", "CALIB_DIR"))
DRS4_PEDESTAL_BASEDIR = Path(cfg.get("LST1", "PEDESTAL_DIR"))


def analysis_path(tel) -> Path:
    """
    Path of the running_analysis directory for a certain date

    Returns
    -------
    directory : Path
        Path of the running_analysis directory for a certain date
    """
    log.debug(f"Getting analysis path for telescope {tel}")
    flat_date = utils.date_to_dir(options.date)
    options.prod_id = utils.get_prod_id()
    directory = Path(cfg.get(tel, "ANALYSIS_DIR")) / flat_date / options.prod_id

    if not options.simulate:
        directory.mkdir(parents=True, exist_ok=True)
    else:
        log.debug("SIMULATE the creation of the analysis directory.")

    log.debug(f"Analysis directory: {directory}")
    return directory


def get_run_date(run_id: int) -> datetime:
    """
    Return the date (YYYYMMDD) when the given run was taken. The search for this date
    is done by looking at the date corresponding to each run in the merged run summaries
    file.
    """
    merged_run_summaries_file = cfg.get("LST1", "MERGED_SUMMARY")
    summary_table = Table.read(merged_run_summaries_file)

    try:
        date_string = summary_table[summary_table["run_id"] == run_id]["date"][0]
    except IndexError:
        log.warning(
            f"Run {run_id} is not in the summary table. "
            f"Assuming the date of the run is {options.date}."
        )
        date_string = utils.date_to_iso(options.date)

    return datetime.strptime(date_string, "%Y-%m-%d")


def get_drs4_pedestal_filename(run_id: int, prod_id: str) -> Path:
    """
    Return the drs4 pedestal file corresponding to a given run id
    regardless of the date when the run was taken.
    """
    if drs4_pedestal_exists(run_id, prod_id):
        files = search_drs4_files(run_id, prod_id)
        return files[-1]  # Get the latest production among the major lstchain version

    date = utils.date_to_dir(get_run_date(run_id))
    return (
        DRS4_PEDESTAL_BASEDIR
        / date
        / f"v{lstchain.__version__}/drs4_pedestal.Run{run_id:05d}.0000.h5"
    ).resolve()


def get_calibration_filename(run_id: int, prod_id: str) -> Path:
    # sourcery skip: remove-unnecessary-cast
    """
    Return the calibration file corresponding to a given run_id.

    Parameters
    ----------
    run_id : int
        Run id of the calibration file to be built.

    Notes
    -----
    The file path will be built regardless of the date when the run was taken.
    We follow the naming convention of the calibration files produced by the
    lstchain script which depends on the filter wheels position. Therefore, we
    need to try to fetch the filter position from the CaCo database. If the
    filter position is not found, we assume the default filter position 5-2.
    Filter information is not available in the database for runs taken before
    mid 2021 approx.
    """

    if calibration_file_exists(run_id, prod_id):
        files = search_calibration_files(run_id, prod_id)
        return files[-1]  # Get the latest production among the major lstchain version

    date = utils.date_to_dir(get_run_date(run_id))

    if options.test:  # Run tests avoiding the access to the database
        options.filters = 52

    else:
        mongodb = cfg.get("database", "caco_db")
        try:
            # Cast run_id to int to avoid problems with numpy int64 encoding in MongoDB
            options.filters = find_filter_wheels(int(run_id), mongodb)
        except IOError:
            log.warning("No filter information found in database. Assuming positions 52.")
            options.filters = 52

    return (
        CALIB_BASEDIR
        / date
        / f"v{lstchain.__version__}/calibration_filters_{options.filters}.Run{run_id:05d}.0000.h5"
    ).resolve()


def pedestal_ids_file_exists(run_id: int) -> bool:
    """Look for the files with pedestal interleaved event identification."""
    pedestal_ids_dir = Path(cfg.get("LST1", "PEDESTAL_FINDER_DIR"))
    file_list = sorted(pedestal_ids_dir.rglob(f"pedestal_ids_Run{run_id:05d}.*.h5"))
    return bool(file_list)


def drs4_pedestal_exists(run_id: int, prod_id: str) -> bool:
    """Return true if drs4 pedestal file was already produced."""
    files = search_drs4_files(run_id, prod_id)

    return len(files) != 0


def calibration_file_exists(run_id: int, prod_id: str) -> bool:
    """Return true if calibration file was already produced."""
    files = search_calibration_files(run_id, prod_id)

    return len(files) != 0


def search_drs4_files(run_id: int, prod_id: str) -> list:
    """
    Find DRS4 baseline correction files corresponding to a run ID
    and major lstchain production version
    """
    date = utils.date_to_dir(get_run_date(run_id))
    version = get_major_version(prod_id)
    drs4_dir = DRS4_PEDESTAL_BASEDIR / date
    return sorted(
        drs4_dir.glob(f"{version}*/drs4_pedestal.Run{run_id:05d}.0000.h5")
    )


def get_major_version(prod_id):
    """Given a version as vX.Y.Z return vX.Y"""
    # First check that the given version is in the correct format
    if prod_id.startswith("v") and len(prod_id.split(".")) >= 2:
        return re.search(r"\D\d+\.\d+", prod_id)[0]

    raise ValueError("Format of the version is not in the form vW.X.Y.Z")


def search_calibration_files(run_id: int, prod_id: str) -> list:
    """
    Search charge calibration files corresponding to a run ID and major lstchain production version
    """
    date = utils.date_to_dir(get_run_date(run_id))
    version = get_major_version(prod_id)
    return sorted(
        (CALIB_BASEDIR / date).glob(f"{version}*/calibration_filters_*.Run{run_id:05d}.0000.h5")
    )


def get_drive_file(date: str) -> Path:
    """Return the drive file corresponding to a given date in YYYYMMDD format."""
    drive_dir = Path(cfg.get("LST1", "DRIVE_DIR"))
    return (drive_dir / f"DrivePosition_log_{date}.txt").resolve()


def get_summary_file(date) -> Path:
    """Return the run summary file corresponding to a given date in YYYYMMDD format."""
    run_summary_dir = Path(cfg.get("LST1", "RUN_SUMMARY_DIR"))
    return (run_summary_dir / f"RunSummary_{date}.ecsv").resolve()


def get_pedestal_ids_file(run_id: int, date: str) -> Path:
    """
    Return the pedestal ids file path corresponding to a given run
    from a date in format YYYYMMDD.
    """
    pedestal_ids_dir = Path(cfg.get("LST1", "PEDESTAL_FINDER_DIR")) / date
    file = pedestal_ids_dir / f"pedestal_ids_Run{run_id:05d}.{{subruns:04d}}.h5"
    return file.resolve()


def sequence_calibration_files(sequence_list: List[Sequence]) -> None:
    """Build names of the calibration files for each sequence in the list."""
    flat_date = utils.date_to_dir(options.date)
    base_dir = Path(cfg.get("LST1", "BASE"))
    prod_id = options.prod_id

    for sequence in sequence_list:
        # Assign the calibration files to the sequence object
        sequence.drs4_file = get_drs4_pedestal_filename(
            sequence.drs4_run, prod_id
        )
        sequence.calibration_file = get_calibration_filename(
            sequence.pedcal_run, prod_id
        )
        sequence.time_calibration_file = find_time_calibration_file(
            "pro", sequence.pedcal_run, base_dir=base_dir
        )
        sequence.systematic_correction_file = find_systematics_correction_file(
            "pro", flat_date, base_dir=base_dir
        )


def get_datacheck_files(pattern: str, directory: Path) -> list:
    """Return a list of files matching the pattern."""
    return sorted(directory.glob(pattern))


def datacheck_directory(data_type: str, date: str) -> Path:
    """Returns the path to the datacheck directory given the data type."""
    if data_type in {"PEDESTAL", "CALIB"}:
        directory = Path(cfg.get("LST1", f"{data_type}_DIR")) / date / "pro/log"
    elif data_type == "DL1AB":
        directory = destination_dir("DATACHECK", create_dir=False)
    elif data_type == "LONGTERM":
        directory = Path(cfg.get("LST1", f"{data_type}_DIR")) / options.prod_id / date
    else:
        raise ValueError(f"Unknown data type: {data_type}")
    return directory


def destination_dir(concept: str, create_dir: bool = True) -> Path:
    """
    Create final destination directory for each data level.
    See Also osa.utils.register_run_concept_files

    Parameters
    ----------
    concept : str
        Expected: MUON, DL1AB, DATACHECK, INTERLEAVED, DL2, PEDESTAL, CALIB, TIMECALIB
    create_dir : bool
        Set it to True (default) if you want to create the directory.
        Otherwise, it just returns the path

    Returns
    -------
    path : pathlib.Path
        Path to the directory
    """
    nightdir = utils.date_to_dir(options.date)

    if concept == "MUON":
        directory = Path(cfg.get(options.tel_id, "DL1_DIR")) / nightdir / options.prod_id / "muons"
    elif concept == "INTERLEAVED":
        directory = (
            Path(cfg.get(options.tel_id, "DL1_DIR")) / nightdir / options.prod_id / "interleaved"
        )
    elif concept == "DATACHECK":
        directory = (
            Path(cfg.get(options.tel_id, "DL1_DIR"))
            / nightdir
            / options.prod_id
            / options.dl1_prod_id
            / "datacheck"
        )
    elif concept == "DL1AB":
        directory = (
            Path(cfg.get(options.tel_id, "DL1_DIR"))
            / nightdir
            / options.prod_id
            / options.dl1_prod_id
        )
    elif concept in {"DL2", "DL3"}:
        directory = (
            (Path(cfg.get(options.tel_id, f"{concept}_DIR")) / nightdir)
            / options.prod_id
        ) / options.dl2_prod_id
    elif concept in {"PEDESTAL", "CALIB", "TIMECALIB"}:
        directory = (
            Path(cfg.get(options.tel_id, f"{concept}_DIR"))
            / nightdir
            / options.prod_id
        )
    elif concept == "HIGH_LEVEL":
        directory = (
            Path(cfg.get(options.tel_id, f"{concept}_DIR"))
            / nightdir
            / options.prod_id
        )
    else:
        log.warning(f"Concept {concept} not known")
        directory = None

    if not options.simulate and create_dir:
        log.debug(f"Destination directory created for {concept}: {directory}")
        directory.mkdir(parents=True, exist_ok=True)
    else:
        log.debug(f"SIMULATING creation of final directory for {concept}")

    return directory


def create_source_directories(source_list: list, cuts_dir: Path):
    """Create a subdirectory for each source."""
    for source in source_list:
        if source is not None:
            source_dir = cuts_dir / source
            source_dir.mkdir(parents=True, exist_ok=True)



"""Identify files to be moved to their final destination directories"""

import logging
import re
import shutil
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.paths import destination_dir
from osa.utils.logging import myLogger
from osa.veto import set_closed_sequence

__all__ = [
    "register_files",
    "register_run_concept_files",
    "register_found_pattern",
    "register_non_existing_file",
]

log = myLogger(logging.getLogger(__name__))

ANALYSIS_PRODUCTS = [
    "DL1AB",
    "DATACHECK",
    "PEDESTAL",
    "CALIB",
    "TIMECALIB",
    "MUON",
    "DL2",
    "INTERLEAVED"
]


def register_files(run_str, analysis_dir, prefix, suffix, output_dir) -> None:
    """
    Copy files into final data directory destination and register
    them into the DB (to be implemented).

    Parameters
    ----------
    run_str: str
        Run number
    analysis_dir: pathlib.Path
        analysis directory
    suffix: str
        suffix of the data file
    output_dir: pathlib.Path
        final data directory
    prefix: str
        prefix of the data file
    """

    file_list = analysis_dir.rglob(f"{prefix}*{run_str}*{suffix}")

    for input_file in file_list:
        output_file = output_dir / input_file.name
        if not output_file.exists():
            log.debug(f"Moving file {input_file} to {output_dir}")
            shutil.move(input_file, output_file)
            # Keep DL1 and muons symlink in running_analysis
            create_symlinks(input_file, output_file, prefix, suffix)


def create_symlinks(input_file, output_file, prefix, suffix):
    """
    Keep DL1 and muons symlink in running_analysis for possible future re-use.
    DL1 symlink is also kept in the DL1ab subdirectory to be able to process
    up to DL2 later on.
    """

    analysis_dir = Path(options.directory)
    dl1ab_dir = analysis_dir / options.dl1_prod_id

    if prefix == "dl1_LST-1" and suffix == ".h5":
        dl1_filepath_analysis_dir = analysis_dir / input_file.name
        dl1_filepath_dl1_dir = dl1ab_dir / input_file.name
        # Remove the original DL1 files pre DL1ab stage and keep only symlinks
        if dl1_filepath_analysis_dir.is_file() and not dl1_filepath_analysis_dir.is_symlink():
            dl1_filepath_analysis_dir.unlink()

        if not dl1_filepath_analysis_dir.is_symlink():
            dl1_filepath_analysis_dir.symlink_to(output_file.resolve())

        # Also set the symlink in the DL1ab subdirectory
        if not dl1_filepath_dl1_dir.is_symlink():
            dl1_filepath_dl1_dir.symlink_to(output_file.resolve())

    if prefix == "muons_LST-1" and suffix == ".fits":
        input_file.symlink_to(output_file.resolve())

    if prefix == "interleaved_LST-1" and suffix == ".h5":
        input_file.symlink_to(output_file.resolve())


def register_run_concept_files(run_string, concept):
    """
    Prepare files to be moved to final destination directories
    from the running_analysis original directory.

    Parameters
    ----------
    run_string: str
    concept: str
    """

    initial_dir = Path(options.directory)  # running_analysis

    # For MUON and INTERLEAVED data products, the initial directory is running_analysis

    if concept == "DL2":
        initial_dir = initial_dir / options.dl2_prod_id

    elif concept == "DL1AB":
        initial_dir = initial_dir / options.dl1_prod_id

    elif concept == "DATACHECK":
        initial_dir = initial_dir / options.dl1_prod_id

    output_dir = destination_dir(concept, create_dir=False)
    data_level = cfg.get("PATTERN", f"{concept}TYPE")
    prefix = cfg.get("PATTERN", f"{concept}PREFIX")
    suffix = cfg.get("PATTERN", f"{concept}SUFFIX")

    log.debug(f"Registering {data_level} file for {prefix}*{run_string}*{suffix}")
    if concept in ANALYSIS_PRODUCTS:
        register_files(run_string, initial_dir, prefix, suffix, output_dir)
    else:
        log.warning(f"Concept {concept} not known")


def register_found_pattern(file_path: Path, seq_list: list, concept: str, destination_path: Path):
    """

    Parameters
    ----------
    file_path: pathlib.Path
    seq_list: list
    concept: str
    destination_path: pathlib.Path
    """
    new_dst = destination_path / file_path.name
    log.debug(f"New file path {new_dst}")
    if not options.simulate:
        if new_dst.exists():
            log.debug("Destination file already exists")
        else:
            log.debug(f"Destination file {new_dst} does not exists")
            register_non_existing_file(file_path, concept, seq_list)

    # Return filepath already registered to be deleted from the set of all files
    return file_path


def register_non_existing_file(file_path, concept, seq_list):
    """

    Parameters
    ----------
    file_path: pathlib.Path
    concept: str
    seq_list: list
    """
    for sequence in seq_list:
        if sequence.type == "DATA":
            run_str_found = re.search(sequence.run_str, str(file_path))

            if run_str_found is not None:
                log.debug(f"Registering file {run_str_found}")
                register_run_concept_files(sequence.run_str, concept)
                if options.seqtoclose is None and not file_path.exists():
                    log.debug("File does not exists")

        elif sequence.type in ["PEDCALIB", "DRS4"]:
            calib_run_str_found = re.search(str(sequence.run), str(file_path))
            drs4_run_str_found = re.search(str(sequence.previousrun), str(file_path))

            if calib_run_str_found is not None:
                log.debug(f"Registering file {calib_run_str_found}")
                register_run_concept_files(str(sequence.run), concept)
                if options.seqtoclose is None and not file_path.exists():
                    log.debug("File does not exists")

            if drs4_run_str_found is not None:
                log.debug(f"Registering file {drs4_run_str_found}")
                register_run_concept_files(str(sequence.previousrun), concept)
                if options.seqtoclose is None and not file_path.exists():
                    log.debug("File does not exists")

        set_closed_sequence(sequence)



import pytest
from pymongo.errors import ConnectionFailure


def test_query():
    from osa.nightsummary import database

    with pytest.raises(ConnectionFailure):
        database.query(obs_id=1616, property_name="DriveControl_SourceName")



"""Extract subrun, run, sequence list and build corresponding objects."""


import itertools
import logging
import sys
from collections import defaultdict
from datetime import datetime, timedelta
from pathlib import Path
from typing import List

import numpy as np
from astropy.table import Table

from osa.configs import options
from osa.configs.config import cfg
from osa.configs.datamodel import (
    RunObj,
    SequenceCalibration,
    SequenceData,
)
from osa.configs.datamodel import Sequence
from osa.job import sequence_filenames
from osa.nightsummary import database
from osa.nightsummary.nightsummary import run_summary_table
from osa.paths import sequence_calibration_files, get_run_date
from osa.utils.logging import myLogger
from osa.utils.utils import date_to_iso, date_to_dir

log = myLogger(logging.getLogger(__name__))

__all__ = [
    "extract_runs",
    "extract_sequences",
    "build_sequences",
    "get_source_list",
]


def get_data_runs(date: datetime):
    """Return the list of DATA runs to analyze based on the run summary table."""
    summary = run_summary_table(date)
    return summary[summary["run_type"] == "DATA"]["run_id"].tolist()


def get_last_drs4(date: datetime) -> int:
    """Return run_id of the last DRS4 run for the given date to be used for data processing."""
    summary = run_summary_table(date)
    n_max = 4
    n = 1
    while (np.array(summary["run_type"] == "DRS4")).any() == False & n <= n_max:
        date = date - timedelta(days=1)
        summary = run_summary_table(date)
        n += 1

    try:
        return summary[summary["run_type"] == "DRS4"]["run_id"].max()

    except ValueError:
        log.warning("No DRS4 run found. Nothing to do. Exiting.")
        sys.exit(0)


def get_last_pedcalib(date) -> int:
    """Return run_id of the last PEDCALIB run for the given date to be used for data processing."""
    summary = run_summary_table(date)
    n_max = 4
    n = 1
    while (np.array(summary["run_type"] == "PEDCALIB")).any() == False & n <= n_max:
        date = date - timedelta(days=1)
        summary = run_summary_table(date)
        n += 1

    try:
        return summary[summary["run_type"] == "PEDCALIB"]["run_id"].max()

    except ValueError:
        log.warning("No PEDCALIB run found. Nothing to do. Exiting.")
        sys.exit(0)


def extract_runs(summary_table):
    """
    Extract sub-wun wise information from RunSummary files.

    Parameters
    ----------
    summary_table: astropy.Table
        Table containing run-wise information indicated in `nightsummary.run_summary`

    See Also: `nightsummary.run_summary`

    Returns
    -------
    subrun_list
    """

    if len(summary_table) == 0:
        log.warning("No runs found for this date. Nothing to do. Exiting.")
        sys.exit(0)

    run_list = []

    required_drs4_run = get_last_drs4(options.date)
    required_pedcal_run = get_last_pedcalib(options.date)

    if required_drs4_run not in summary_table["run_id"]:
        drs4_date = get_run_date(required_drs4_run)
        drs4_date_summary = run_summary_table(drs4_date)
        run_info = drs4_date_summary.loc[required_drs4_run]
        run = RunObj(
            run=required_drs4_run,
            run_str=f"{required_drs4_run:05d}",
            type=run_info["run_type"],
            night=date_to_iso(drs4_date),
            subruns=run_info["n_subruns"],
        )
        run_list.append(run)

    if required_pedcal_run not in summary_table["run_id"]:
        pedcal_date = get_run_date(required_pedcal_run)
        pedcal_date_summary = run_summary_table(pedcal_date)
        run_info = pedcal_date_summary.loc[required_pedcal_run]
        run = RunObj(
            run=required_pedcal_run,
            run_str=f"{required_pedcal_run:05d}",
            type=run_info["run_type"],
            night=date_to_iso(pedcal_date),
            subruns=run_info["n_subruns"],
        )
        run_list.append(run)

    # Get information run-wise going through each row
    for run_id in summary_table["run_id"]:
        run_info = summary_table.loc[run_id]
        # Build run object
        run = RunObj(
            run=run_id,
            run_str=f"{run_id:05d}",
            type=run_info["run_type"],
            night=date_to_iso(options.date),
            subruns=run_info["n_subruns"],
        )
        run_list.append(run)

    # Before trying to access the information in the database, check if it is available
    # in the RunCatalog file (previously generated with the information in the database).
    nightdir = date_to_dir(options.date)
    source_catalog_dir = Path(cfg.get("LST1", "RUN_CATALOG"))
    source_catalog_dir.mkdir(parents=True, exist_ok=True)
    source_catalog_file = source_catalog_dir / f"RunCatalog_{nightdir}.ecsv"

    if source_catalog_file.exists():
        log.debug(f"RunCatalog file found: {source_catalog_file}")
        source_catalog = Table.read(source_catalog_file)
        # Add index to be able to browse the table
        source_catalog.add_index("run_id")

        # Get information run-wise going through each row of the RunCatalog file
        # and assign it to the corresponding run object.
        for run_id, run in itertools.product(source_catalog["run_id"], run_list):
            if run.run == run_id:
                run.source_name = source_catalog.loc[run_id]["source_name"]
                run.source_ra = source_catalog.loc[run_id]["source_ra"]
                run.source_dec = source_catalog.loc[run_id]["source_dec"]

    elif database.db_available():
        run_table = Table(
            names=["run_id", "source_name", "source_ra", "source_dec"],
            dtype=["int32", str, "float64", "float64"],
        )
        for run in run_list:
            # Make sure we are looking at actual data runs. Avoid test runs.
            if run.run > 0 and run.type == "DATA":
                log.debug(f"Looking info in TCU DB for run {run.run}")
                run.source_name = database.query(
                    obs_id=run.run, property_name="DriveControl_SourceName"
                )
                run.source_ra = database.query(
                    obs_id=run.run, property_name="DriveControl_RA_Target"
                    )
                run.source_dec = database.query(
                    obs_id=run.run, property_name="DriveControl_Dec_Target"
                    )
                # Store this source information (run_id, source_name, source_ra, source_dec)
                # into an astropy Table and save to disk in RunCatalog files. In this way, the
                # information can be dumped anytime later more easily than accessing the
                # TCU database.
                if run.source_name is not None:
                    line = [
                        run.run,
                        run.source_name,
                        run.source_ra,
                        run.source_dec,
                    ]
                    log.debug(f"Adding line with source info to RunCatalog: {line}")
                    run_table.add_row(line)

        # Save table to disk
        run_table.write(source_catalog_file, overwrite=True, delimiter=",")

    log.debug("Subrun list extracted")

    return run_list


def extract_sequences(date: datetime, run_obj_list: List[RunObj]) -> List[Sequence]:
    """
    Create calibration and data sequences from run objects.

    Parameters
    ----------
    date : datetime
        Date of the runs to analyze
    run_obj_list : List[RunObj]
        List of run objects

    Returns
    -------
    sequence_list : List[Sequence]
    """

    # Last DRS4 and PEDCALIB runs required to process the sky-data runs
    required_drs4_run = get_last_drs4(date)
    required_pedcal_run = get_last_pedcalib(date)

    # Get DATA runs to be processed
    data_runs_to_process = get_data_runs(date)
    if not data_runs_to_process:
        log.warning("No data sequences found for this date. Nothing to do. Exiting.")
        sys.exit(0)

    log.debug(f"There are {len(data_runs_to_process)} data sequences: {data_runs_to_process}")

    # Loop over the list of run objects and create the corresponding sequences
    # First, the calibration sequence based on the last DRS4 and PEDCALIB runs.
    # Then, the data sequences which require the calibration files produced
    # by the calibration sequence.
    sequence_list = []

    for run in run_obj_list:
        if run.run == required_pedcal_run:
            sequence = SequenceCalibration(run)
            sequence.jobname = f"{run.telescope}_{run.run:05d}"
            sequence_list.insert(0, sequence)
            sequence.drs4_run = required_drs4_run
            sequence.pedcal_run = run.run
            sequence_filenames(sequence)
            log.debug(
                f"Calibration sequence {sequence.seq} composed of "
                f"DRS4 run {required_drs4_run} and Ped-Cal run {required_pedcal_run}"
            )

        elif run.run in data_runs_to_process:
            sequence = SequenceData(run)
            # data sequences counted after the calibration sequence
            sequence.seq = data_runs_to_process.index(run.run) + 2
            sequence.jobname = f"{run.telescope}_{run.run:05d}"
            sequence.drs4_run = required_drs4_run
            sequence.pedcal_run = required_pedcal_run
            sequence_filenames(sequence)
            log.debug(
                f"Data sequence {sequence.seq} from run {run.run} whose parent is "
                f"{sequence.parent} (DRS4 {required_drs4_run} & Ped-Cal {required_pedcal_run})"
            )
            sequence_list.append(sequence)

    # Add the calibration file names
    sequence_calibration_files(sequence_list)
    log.debug("Workflow completed")

    log.debug("Sequence list extracted")

    return sequence_list


def build_sequences(date: datetime) -> List:
    """Build the list of sequences to process from a given date."""
    summary_table = run_summary_table(date)
    run_list = extract_runs(summary_table)
    # modifies run_list by adding the seq and parent info into runs
    return extract_sequences(date, run_list)


def get_source_list(date: datetime) -> dict:
    """
    Get the list of sources from the sequences' information and corresponding runs.

    Parameters
    ----------
    date : datetime

    Returns
    -------
    sources : Dict[str, list]
    """

    # Build the sequences
    sequence_list = build_sequences(date)

    # Create a dictionary of sources and their corresponding sequences
    source_dict = {
        sequence.run: sequence.source_name
        for sequence in sequence_list
        if sequence.source_name is not None
    }

    source_dict_grouped = defaultdict(list)
    for key, val in sorted(source_dict.items()):
        source_dict_grouped[val].append(key)

    if list(source_dict_grouped) is None:
        sys.exit("No sources found. Check the access to database. Exiting.")

    return dict(source_dict_grouped)



"""
Display summary of the observations for a given date.

Show the run summary for a given date containing the number of subruns,
the start and end time of the run and type pf the run: DATA, DRS4, PEDCALIB.
"""

import argparse
import logging
import os
from pathlib import Path

import numpy as np
from astropy import units as u
from astropy.table import Table
from astropy.time import Time
from lstchain.scripts.lstchain_create_run_summary import (
    get_list_of_files,
    get_list_of_runs,
    get_runs_and_subruns,
    type_of_run,
    read_counters,
)

from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))

parser = argparse.ArgumentParser(description="Create run summary file")

parser.add_argument(
    "-d",
    "--date",
    help="Date for the creation of the run summary in format YYYYMMDD",
    required=True,
)

parser.add_argument(
    "--r0-path",
    type=Path,
    help="Path to the R0 files. Default is /fefs/aswg/data/real/R0",
    default=Path("/fefs/aswg/data/real/R0"),
)

dtypes = {
    "time_start": str,
    "time_end": str,
    "elapsed": u.quantity.Quantity,
}


def start_end_of_run_files_stat(r0_path: Path, run_number: int, num_files: int):
    """
    Get first timestamps from the last subrun.
    Write down the reference Dragon module used, reference event_id.

    Notes
    -----
    Start and end times are currently taken from the creation and last modification
    time of the first and last file in the run. They are approximate and may be off
    by a few seconds.

    Parameters
    ----------
    r0_path : pathlib.Path
        Directory that contains the R0 files
    run_number : int
        Number of the run
    num_files : int
        Number of the sequential files (subruns) of a given run

    Returns
    -------
    end_timestamp
    """

    last_subrun = num_files - 1  # first subrun is 0
    pattern_first_subrun = r0_path / f"LST-1.1.Run{run_number:05d}.0000.fits.fz"
    pattern_last_subrun = r0_path / f"LST-1.1.Run{run_number:05d}.{last_subrun:04d}.fits.fz"
    try:
        # Get start and end times from the creation and last modification timestamps
        # from the first and last file in the run
        run_start = Time(os.path.getctime(pattern_first_subrun), format="unix")
        run_end = Time(os.path.getmtime(pattern_last_subrun), format="unix")
        elapsed_time = run_end - run_start

        return dict(
            time_start=run_start.iso,
            time_end=run_end.iso,
            elapsed=np.round(elapsed_time.to_value("min"), decimals=1),
        )

    except Exception as err:
        log.error(f"Files {pattern_first_subrun} or {pattern_last_subrun} have error: {err}")

        return dict(
            time_start=None,
            time_end=None,
            elapsed=0.0,
        )


def main():
    """
    Build an astropy Table with run summary information and write it
    as ECSV file with the following information (one row per run):
     - run_id
     - number of subruns
     - type of run (DRS4, CALI, DATA, CONF)
     - start of the run
     - dragon reference UCTS timestamp if available (-1 otherwise)
     - dragon reference time source ("ucts" or "run_date")
     - dragon_reference_module_id
     - dragon_reference_module_index
     - dragon_reference_counter
    """

    log.setLevel(logging.INFO)

    args = parser.parse_args()

    date_path = args.r0_path / args.date

    file_list = get_list_of_files(date_path)
    runs = get_list_of_runs(file_list)
    run_numbers, n_subruns = get_runs_and_subruns(runs)

    reference_counters = [read_counters(date_path, run) for run in run_numbers]

    run_types = [
        type_of_run(date_path, run, counters)
        for run, counters in zip(run_numbers, reference_counters)
    ]

    start_end_timestamps = [
        start_end_of_run_files_stat(date_path, run, n_files)
        for run, n_files in zip(run_numbers, n_subruns)
    ]

    run_summary = Table(
        {
            col: np.array([d[col] for d in start_end_timestamps], dtype=dtype)
            for col, dtype in dtypes.items()
        }
    )

    run_summary.add_column(run_numbers, name="run_id", index=0)
    run_summary.add_column(n_subruns, name="n_subruns", index=1)
    run_summary.add_column(run_types, name="run_type", index=2)

    run_summary["elapsed"].unit = u.min

    header = " Run summary "
    print(f"{header.center(50, '*')}")
    run_summary.pprint_all()
    print("\n")

    # Sum elapsed times:
    obs_by_type = run_summary.group_by("run_type")
    obs_by_type["number_of_runs"] = 1
    total_obs_time = obs_by_type[
        "run_type", "number_of_runs", "n_subruns", "elapsed"
    ].groups.aggregate(np.sum)
    total_obs_time["elapsed"].format = "7.1f"

    header = " Observation time per run type "
    print(f"{header.center(50, '*')}")
    total_obs_time.pprint_all()
    print("\n")

    run_summary["number_of_runs"] = 1
    total_obs = run_summary["number_of_runs", "n_subruns", "elapsed"].groups.aggregate(np.sum)
    total_obs["elapsed"].format = "7.1f"
    header = " Total observation time "
    print(f"{header.center(50, '*')}")
    total_obs.pprint_all()


if __name__ == "__main__":
    main()



"""
Print out basic run summary for a given date with run numbers, type of runs,
start and stop timestamps and elapsed times.
"""

import argparse
import datetime
from pathlib import Path

import astropy.units as u
import numpy as np
from astropy.table import Table
from lstchain.scripts.lstchain_create_run_summary import get_list_of_files, get_list_of_runs

from osa.nightsummary.database import get_run_info_from_TCU

parser = argparse.ArgumentParser(description="Create run summary file")

parser.add_argument(
    "-d",
    "--date",
    help="Date for the creation of the run summary in format YYYYMMDD",
    required=True,
)

parser.add_argument(
    "--r0-path",
    type=Path,
    help="Path to the R0 files. Default is /fefs/aswg/data/real/R0",
    default=Path("/fefs/aswg/data/real/R0"),
)

parser.add_argument(
    "--tcu-db",
    type=str,
    help="Server of the TCU monitoring database",
    default="lst101-int",
)


def main():
    """
    Get run metadata information from TCU monitoring
    database and print out the run summary
    """
    args = parser.parse_args()

    tcu_db = args.tcu_db
    date_path = args.r0_path / args.date
    file_list = get_list_of_files(date_path)
    all_runs = get_list_of_runs(file_list)
    run_numbers = [x.run for x in all_runs]
    run_numbers_array = np.unique(run_numbers)
    run_numbers_array = run_numbers_array[run_numbers_array != 0]

    list_info = []

    for run in run_numbers_array:
        run_info = get_run_info_from_TCU(int(run), tcu_server=tcu_db)
        list_info.append(run_info)

    if list_info:
        table = Table(
            np.array(list_info).T.tolist(),
            names=("run", "type", "tstart", "tstop", "elapsed"),
            dtype=(int, str, datetime.datetime, datetime.datetime, float),
        )
        table["elapsed"].unit = u.min
        table["elapsed"].info.format = "3.1f"
        print("\n")
        table.pprint_all()

        # Sum elapsed times:
        obs_by_type = table.group_by("type")
        obs_by_type["number_of_runs"] = 1
        total_obs_time = obs_by_type["type", "number_of_runs", "elapsed"].groups.aggregate(np.sum)
        total_obs_time["elapsed"].info.format = "7.0f"

        print("\n")
        header = " Observation time per run type "
        print(f"{header.center(50, '*')}")
        total_obs_time.pprint_all()
        print("\n")

    else:
        print(f"No data found in {date_path}")


if __name__ == "__main__":
    main()
